#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import uuid

import octobot_trading.enums as enums
import octobot_trading.personal_data.transactions.transaction as transaction


class RealisedPnlTransaction(transaction.Transaction):
    def __init__(self, exchange_name, creation_time, transaction_type, currency, symbol, realised_pnl, closed_quantity,
                 cumulated_closed_quantity, first_entry_time, average_entry_price, average_exit_price, order_exit_price,
                 leverage, trigger_source, side):
        super().__init__(exchange_name, creation_time, transaction_type, currency, symbol=symbol)
        self.realised_pnl = realised_pnl
        self.closed_quantity = closed_quantity
        self.cumulated_closed_quantity = cumulated_closed_quantity
        self.first_entry_time = first_entry_time
        self.average_entry_price = average_entry_price
        self.average_exit_price = average_exit_price
        self.order_exit_price = order_exit_price
        self.leverage = leverage
        self.trigger_source = trigger_source
        self.side = side
        self.transaction_id = f"{self.exchange_name}" \
                              f"-{str(uuid.uuid4())}" \
                              f"-{self.symbol}" \
                              f"-{str(self.creation_time)}" \
                              f"{'-closed' if self.is_closed_pnl() else ''}"

    def is_closed_pnl(self):
        return self.transaction_type is enums.TransactionType.CLOSE_REALISED_PNL



# pylint: disable=E0611
#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import octobot_backtesting.api as api
import octobot_commons.errors as errors

import octobot_trading.exchange_data.funding.channel.funding_updater as funding_updater
import octobot_trading.util as util


class FundingUpdaterSimulator(funding_updater.FundingUpdater):
    """
    The Funding Update Simulator simulates the exchange funding rate and send it to the Funding Channel
    """
    def __init__(self, channel, importer):
        super().__init__(channel)
        self.exchange_data_importer = importer

        self.initial_timestamp = api.get_backtesting_current_time(self.channel.exchange_manager.exchange.backtesting)
        self.last_pushed_timestamp = 0
        self.time_consumer = None

    async def start(self):
        await self.resume()

    async def handle_timestamp(self, timestamp, **kwargs):
        try:
            current_time = api.get_backtesting_current_time(self.channel.exchange_manager.exchange.backtesting)
            if current_time - self.last_pushed_timestamp > self.FUNDING_REFRESH_TIME_MAX \
                    or self.last_pushed_timestamp == 0:
                self.last_pushed_timestamp = current_time
                for pair in self.channel.exchange_manager.exchange_config.traded_symbol_pairs:
                    funding_rate = await self.channel.exchange_manager.exchange.get_funding_rate(symbol=pair)
                    await self._push_funding(
                        symbol=pair,
                        funding_rate=funding_rate,
                        predicted_funding_rate=funding_rate)
        except errors.DatabaseNotFoundError as e:
            self.logger.warning(f"Not enough data : {e}")
            await self.pause()
            await self.stop()
        except IndexError as e:
            self.logger.warning(f"Failed to access funding_data : {e}")
        except Exception as e:
            self.logger.exception(e, True, f"Error when updating from timestamp: {e}")

    async def pause(self):
        await util.pause_time_consumer(self)

    async def stop(self):
        await util.stop_and_pause(self)

    async def resume(self):
        await util.resume_time_consumer(self, self.handle_timestamp)



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import decimal
from typing import Tuple

import pytest

from octobot_commons.asyncio_tools import wait_asyncio_next_cycle

from octobot_trading.exchange_data.prices.price_events_manager import PriceEventsManager
from octobot_trading.enums import TradeOrderType, TradeOrderSide, MarkPriceSources
from octobot_trading.personal_data.orders import TrailingStopOrder
import octobot_trading.constants as trading_constants
from tests.personal_data import DEFAULT_SYMBOL_QUANTITY, DEFAULT_ORDER_SYMBOL, DEFAULT_MARKET_QUANTITY

from tests.test_utils.random_numbers import decimal_random_price, \
    decimal_random_quantity, decimal_random_recent_trade
from tests import event_loop
from tests.exchanges import simulated_trader, simulated_exchange_manager
from tests.personal_data.orders import trailing_stop_order

pytestmark = pytest.mark.asyncio


async def test_trailing_stop_trigger(trailing_stop_order):
    trailing_stop_order, order_price, price_events_manager = await initialize_trailing_stop(trailing_stop_order)
    await trailing_stop_order.set_trailing_percent(decimal.Decimal(10))
    max_trailing_hit_price = get_price_percent(order_price, trailing_stop_order.trailing_percent)

    # set mark price
    set_mark_price(trailing_stop_order, order_price)

    price_events_manager.handle_recent_trades(
        [decimal_random_recent_trade(
            price=decimal_random_price(min_value=decimal.Decimal(max_trailing_hit_price),
                                       max_value=order_price - trading_constants.ONE),
            timestamp=trailing_stop_order.timestamp)])
    await wait_asyncio_next_cycle()
    assert not trailing_stop_order.is_filled()
    price_events_manager.handle_recent_trades(
        [decimal_random_recent_trade(price=order_price,
                                     timestamp=trailing_stop_order.timestamp - 1)])
    await wait_asyncio_next_cycle()
    assert not trailing_stop_order.is_filled()
    price_events_manager.handle_recent_trades([decimal_random_recent_trade(price=order_price,
                                                                           timestamp=trailing_stop_order.timestamp)])

    await wait_asyncio_next_cycle()
    assert not trailing_stop_order.is_filled()
    # avoid decimal to float rounding issues
    price_events_manager.handle_recent_trades(
        [decimal_random_recent_trade(price=max_trailing_hit_price - decimal.Decimal("0.0001"),
                                     timestamp=trailing_stop_order.timestamp)])
    await wait_asyncio_next_cycle()
    assert trailing_stop_order.is_filled()


async def test_trailing_stop_with_new_price(trailing_stop_order):
    trailing_stop_order, order_price, price_events_manager = await initialize_trailing_stop(trailing_stop_order)
    await trailing_stop_order.set_trailing_percent(decimal.Decimal(2))
    new_trailing_price = decimal_random_price(min_value=order_price * decimal.Decimal(str(0.99)),
                                              max_value=order_price * decimal.Decimal(str(1.01)))

    # set mark price
    set_mark_price(trailing_stop_order, new_trailing_price)

    # move trailing price
    price_events_manager.handle_recent_trades(
        [decimal_random_recent_trade(price=new_trailing_price,
                                     timestamp=trailing_stop_order.timestamp)])
    await wait_asyncio_next_cycle()
    assert not trailing_stop_order.is_filled()

    # test fill stop loss with new order price reference
    # avoid decimal to float rounding issues
    price_events_manager.handle_recent_trades(
        [decimal_random_recent_trade(price=get_price_percent(new_trailing_price - decimal.Decimal("0.0001"),
                                                             trailing_stop_order.trailing_percent),
                                     timestamp=trailing_stop_order.timestamp)])
    await wait_asyncio_next_cycle()
    assert trailing_stop_order.is_filled()


async def test_trailing_stop_with_new_old_price(trailing_stop_order):
    trailing_stop_order, order_price, price_events_manager = await initialize_trailing_stop(trailing_stop_order)
    await trailing_stop_order.set_trailing_percent(decimal.Decimal(5))
    new_trailing_price = decimal_random_price(min_value=order_price * decimal.Decimal(str(0.99)),
                                              max_value=order_price * decimal.Decimal(str(1.01)))

    # set mark price
    set_mark_price(trailing_stop_order, new_trailing_price)

    # move trailing price
    price_events_manager.handle_recent_trades(
        [decimal_random_recent_trade(price=new_trailing_price,
                                     timestamp=trailing_stop_order.timestamp)])
    await wait_asyncio_next_cycle()
    assert not trailing_stop_order.is_filled()

    # test fill stop loss with old price
    price_events_manager.handle_recent_trades(
        [decimal_random_recent_trade(price=get_price_percent(order_price - decimal.Decimal("0.0001"),
                                                             trailing_stop_order.trailing_percent),
                                     timestamp=trailing_stop_order.timestamp)])
    await wait_asyncio_next_cycle()
    assert trailing_stop_order.is_filled()


async def test_trailing_stop_with_new_price_inversed(trailing_stop_order):
    trailing_stop_order, order_price, price_events_manager = await initialize_trailing_stop(trailing_stop_order,
                                                                                            side=TradeOrderSide.BUY)
    await trailing_stop_order.set_trailing_percent(decimal.Decimal(5))
    new_trailing_price = decimal_random_price(min_value=order_price * decimal.Decimal(str(0.99)),
                                              max_value=order_price * decimal.Decimal(str(1.01)))

    # set mark price
    set_mark_price(trailing_stop_order, new_trailing_price)

    # move trailing price
    price_events_manager.handle_recent_trades(
        [decimal_random_recent_trade(price=new_trailing_price,
                                     timestamp=trailing_stop_order.timestamp)])
    await wait_asyncio_next_cycle()
    assert not trailing_stop_order.is_filled()

    # test fill stop loss with new order price reference
    price_events_manager.handle_recent_trades(
        [decimal_random_recent_trade(price=get_price_percent(new_trailing_price + decimal.Decimal("0.0001"),
                                                             trailing_stop_order.trailing_percent,
                                                             selling_side=False),
                                     timestamp=trailing_stop_order.timestamp)])
    await wait_asyncio_next_cycle()
    assert trailing_stop_order.is_filled()


async def initialize_trailing_stop(order, side=TradeOrderSide.SELL) -> Tuple[
    TrailingStopOrder, decimal.Decimal, PriceEventsManager]:
    order_price = decimal_random_price()
    order_max_quantity = DEFAULT_SYMBOL_QUANTITY \
        if side is TradeOrderSide.SELL else DEFAULT_MARKET_QUANTITY / order_price
    order.update(
        price=order_price,
        # divide quantity by 2 to prevent trailing price movement to impact usable quantity
        quantity=decimal_random_quantity(max_value=order_max_quantity / decimal.Decimal(2)),
        symbol=DEFAULT_ORDER_SYMBOL,
        order_type=TradeOrderType.TRAILING_STOP,
    )
    order.side = side
    order.exchange_manager.is_backtesting = True  # force update_order_status
    await order.initialize()
    price_events_manager = order.exchange_manager.exchange_symbols_data.get_exchange_symbol_data(
        DEFAULT_ORDER_SYMBOL).price_events_manager
    await order.exchange_manager.exchange_personal_data.orders_manager.upsert_order_instance(order)
    return order, order_price, price_events_manager


def get_price_percent(price, percent, selling_side=True):
    return price * (1 + (percent / 100) * (-1 if selling_side else 1))


def set_mark_price(order, mark_price):
    prices_manager = order.exchange_manager.exchange_symbols_data. \
        get_exchange_symbol_data(order.symbol).prices_manager
    prices_manager.set_mark_price(decimal.Decimal(mark_price), MarkPriceSources.EXCHANGE_MARK_PRICE.value)
    prices_manager.mark_price_set_time = order.timestamp



#  Drakkar-Software OctoBot
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import time
import importlib

import octobot_commons.logging as logging
import octobot_commons.enums as commons_enums
import octobot_commons.errors as commons_errors
import octobot_commons.constants as commons_constants
import octobot_commons.databases as databases
import octobot_trading.modes.abstract_trading_mode as abstract_trading_mode
import octobot_trading.modes.channel as modes_channel
import octobot_trading.modes.script_keywords.context_management as context_management
import octobot_trading.modes.script_keywords.basic_keywords as basic_keywords
import octobot_trading.modes.modes_util as modes_util
import octobot_trading.errors as errors
import octobot_tentacles_manager.api as tentacles_manager_api


class AbstractScriptedTradingMode(abstract_trading_mode.AbstractTradingMode):
    TRADING_SCRIPT_MODULE = None
    BACKTESTING_SCRIPT_MODULE = None
    ALLOW_CUSTOM_TRIGGER_SOURCE = True

    INITIALIZED_TRADING_PAIR_BY_BOT_ID = {}

    def __init__(self, config, exchange_manager):
        super().__init__(config, exchange_manager)
        self._live_script = None
        self._backtesting_script = None
        self.timestamp = time.time()
        self.script_name = None

        if exchange_manager:
            # add config folder to importable files to import the user script
            tentacles_manager_api.import_user_tentacles_config_folder(self.exchange_manager.tentacles_setup_config)

    def get_current_state(self) -> (str, float):
        return super().get_current_state()[0] if self.producers[0].state is None else self.producers[0].state.name, \
               "N/A"

    def get_mode_producer_classes(self) -> list:
        return [AbstractScriptedTradingModeProducer]

    async def user_commands_callback(self, bot_id, subject, action, data) -> None:
        # do not call super as reload_config is called by reload_scripts already
        # on RELOAD_CONFIG command
        if action == commons_enums.UserCommands.RELOAD_CONFIG.value:
            # also reload script on RELOAD_CONFIG
            await self.reload_scripts()
        elif action == commons_enums.UserCommands.RELOAD_SCRIPT.value:
            await self.reload_scripts()
        elif action == commons_enums.UserCommands.CLEAR_PLOTTING_CACHE.value:
            await modes_util.clear_plotting_cache(self)
        elif action == commons_enums.UserCommands.CLEAR_SIMULATED_ORDERS_CACHE.value:
            await modes_util.clear_simulated_orders_cache(self)

    @classmethod
    async def get_backtesting_plot(cls, exchange, symbol, backtesting_id, optimizer_id,
                                   optimization_campaign, backtesting_analysis_settings):
        ctx = context_management.Context.minimal(cls, logging.get_logger(cls.get_name()), exchange, symbol,
                                                 backtesting_id, optimizer_id,
                                                 optimization_campaign, backtesting_analysis_settings)
        return await cls.get_script_from_module(cls.BACKTESTING_SCRIPT_MODULE)(ctx)

    @classmethod
    def get_is_symbol_wildcard(cls) -> bool:
        return False

    def get_script(self, live=True):
        return self._live_script if live else self._backtesting_script

    def register_script_module(self, script_module, live=True):
        if live:
            self.__class__.TRADING_SCRIPT_MODULE = script_module
            self._live_script = self.get_script_from_module(script_module)
        else:
            self.__class__.BACKTESTING_SCRIPT_MODULE = script_module
            self._backtesting_script = self.get_script_from_module(script_module)

    @staticmethod
    def get_script_from_module(module):
        return module.script

    async def reload_scripts(self):
        for is_live in (False, True):
            module = self.__class__.TRADING_SCRIPT_MODULE if is_live else self.__class__.BACKTESTING_SCRIPT_MODULE
            importlib.reload(module)
            self.register_script_module(module, live=is_live)
            # reload config
            await self.reload_config(self.exchange_manager.bot_id)
            if is_live:
                # todo cancel and restart live tasks
                await self.start_over_database()

    async def start_over_database(self):
        await modes_util.clear_plotting_cache(self)
        symbol_db = databases.RunDatabasesProvider.instance().get_symbol_db(self.bot_id,
                                                                            self.exchange_manager.exchange_name,
                                                                            self.symbol)
        symbol_db.set_initialized_flags(False)
        for producer in self.producers:
            for time_frame, call_args in producer.last_call_by_timeframe.items():
                run_db = databases.RunDatabasesProvider.instance().get_run_db(self.bot_id)
                await producer.init_user_inputs(False)
                run_db.set_initialized_flags(False, (time_frame, ))
                await databases.CacheManager().close_cache(commons_constants.UNPROVIDED_CACHE_IDENTIFIER,
                                                           reset_cache_db_ids=True)
                await producer.call_script(*call_args)
                await run_db.flush()

    def set_initialized_trading_pair_by_bot_id(self, symbol, time_frame, initialized):
        # todo migrate to event tree
        try:
            self.__class__.INITIALIZED_TRADING_PAIR_BY_BOT_ID[self.bot_id][self.exchange_manager.exchange_name][
                symbol][time_frame] = initialized
        except KeyError:
            if self.bot_id not in self.__class__.INITIALIZED_TRADING_PAIR_BY_BOT_ID:
                self.__class__.INITIALIZED_TRADING_PAIR_BY_BOT_ID[self.bot_id] = {}
            if self.exchange_manager.exchange_name not in \
                    self.__class__.INITIALIZED_TRADING_PAIR_BY_BOT_ID[self.bot_id]:
                self.__class__.INITIALIZED_TRADING_PAIR_BY_BOT_ID[self.bot_id][self.exchange_manager.exchange_name] = {}
            if symbol not in \
                    self.__class__.INITIALIZED_TRADING_PAIR_BY_BOT_ID[self.bot_id][self.exchange_manager.exchange_name]:
                self.__class__.INITIALIZED_TRADING_PAIR_BY_BOT_ID[self.bot_id][
                    self.exchange_manager.exchange_name][symbol] = {}
            if time_frame not in \
                    self.__class__.INITIALIZED_TRADING_PAIR_BY_BOT_ID[self.bot_id][self.exchange_manager.exchange_name][
                        symbol]:
                self.__class__.INITIALIZED_TRADING_PAIR_BY_BOT_ID[self.bot_id][self.exchange_manager.exchange_name][
                    symbol][time_frame] = initialized

    def get_initialized_trading_pair_by_bot_id(self, symbol, time_frame):
        try:
            return self.__class__.INITIALIZED_TRADING_PAIR_BY_BOT_ID[self.bot_id][self.exchange_manager.exchange_name][
                    symbol][time_frame]
        except KeyError:
            return False

    async def get_additional_metadata(self, _):
        return {
            commons_enums.BacktestingMetadata.NAME.value: self.script_name
        }


class AbstractScriptedTradingModeProducer(modes_channel.AbstractTradingModeProducer):

    def __init__(self, channel, config, trading_mode, exchange_manager):
        super().__init__(channel, config, trading_mode, exchange_manager)
        self.last_call_by_timeframe = {}

    async def start(self) -> None:
        await super().start()
        # refresh user inputs
        if not self.exchange_manager.is_backtesting:
            await self._schedule_initialization_call()

    async def _schedule_initialization_call(self):
        # initialization call is a special call that does not trigger trades and allows the script
        # to be run at least once in order to initialize its configuration
        if self.exchange_manager.is_backtesting:
            # not necessary in backtesting
            return

        # fake a full candle call
        cryptocurrency, symbol, time_frame = self._get_initialization_call_args()
        # wait for symbol data to be initialized
        candle = await self._wait_for_symbol_init(symbol, time_frame, self.CONFIG_INIT_TIMEOUT)
        if candle is None:
            self.logger.error(f"Can't initialize trading script: {symbol} {time_frame} candles are not fetched")
        await self.ohlcv_callback(self.exchange_name, self.exchange_manager.id, cryptocurrency, symbol, time_frame,
                                  candle, init_call=True)

    async def _wait_for_symbol_init(self, symbol, time_frame, timeout):
        if not await super()._wait_for_symbol_init(symbol, time_frame, timeout):
            return None
        return self._get_latest_full_candle(symbol, time_frame)

    def _get_latest_full_candle(self, symbol, time_frame):
        tf = commons_enums.TimeFrames(time_frame)
        candles_manager = self.exchange_manager.exchange_symbols_data.get_exchange_symbol_data(
            symbol,
            allow_creation=False) \
            .symbol_candles[tf]
        candle_data = candles_manager.get_candles(5)
        current_time = self.exchange_manager.exchange.get_exchange_current_time()
        time_frame_sec = commons_enums.TimeFramesMinutes[tf] * commons_constants.MINUTE_TO_SECONDS
        last_full_candle_time = current_time - current_time % time_frame_sec - time_frame_sec
        for candle in reversed(candle_data):
            if candle[commons_enums.PriceIndexes.IND_PRICE_TIME.value] == last_full_candle_time:
                return candle
        # return the candle right before the last (last being in construction)
        return candle_data[-2]

    def _get_initialization_call_args(self):
        currency = next(iter(self.exchange_manager.exchange_config.traded_cryptocurrencies))
        symbol = self.exchange_manager.exchange_config.traded_cryptocurrencies[currency][0]
        time_frame = self.exchange_manager.exchange_config.traded_time_frames[0]
        return currency, symbol, time_frame.value

    async def ohlcv_callback(self, exchange: str, exchange_id: str, cryptocurrency: str, symbol: str,
                             time_frame: str, candle: dict, init_call: bool = False):
        async with self.trading_mode_trigger(), self.trading_mode.remote_signal_publisher(symbol):
            # add a full candle to time to get the real time
            trigger_time = candle[commons_enums.PriceIndexes.IND_PRICE_TIME.value] + \
                           commons_enums.TimeFramesMinutes[commons_enums.TimeFrames(time_frame)] * \
                           commons_constants.MINUTE_TO_SECONDS
            await self.call_script(self.matrix_id, cryptocurrency, symbol, time_frame,
                                   commons_enums.TriggerSource.OHLCV.value,
                                   trigger_time,
                                   candle=candle,
                                   init_call=init_call)

    async def kline_callback(self, exchange: str, exchange_id: str, cryptocurrency: str, symbol: str,
                             time_frame, kline: dict):
        async with self.trading_mode_trigger(), self.trading_mode.remote_signal_publisher(symbol):
            await self.call_script(self.matrix_id, cryptocurrency, symbol, time_frame,
                                   commons_enums.TriggerSource.KLINE.value,
                                   kline[commons_enums.PriceIndexes.IND_PRICE_TIME.value],
                                   kline=kline)

    async def set_final_eval(self, matrix_id: str, cryptocurrency: str, symbol: str, time_frame, trigger_source: str):
        await self.call_script(matrix_id, cryptocurrency, symbol, time_frame, trigger_source,
                               self._get_latest_eval_time(matrix_id, cryptocurrency, symbol, time_frame))

    def _get_latest_eval_time(self, matrix_id: str, cryptocurrency: str, symbol: str, time_frame):
        try:
            import octobot_evaluators.matrix as matrix
            import octobot_evaluators.enums as evaluators_enums
            return matrix.get_latest_eval_time(matrix_id,
                                               exchange_name=self.exchange_name,
                                               tentacle_type=evaluators_enums.EvaluatorMatrixTypes.SCRIPTED.value,
                                               cryptocurrency=cryptocurrency,
                                               symbol=symbol,
                                               time_frame=time_frame)
        except ImportError:
            self.logger.error("OctoBot-Evaluators is required for a matrix callback")
            return None

    async def call_script(self, matrix_id: str, cryptocurrency: str, symbol: str, time_frame: str,
                          trigger_source: str, trigger_cache_timestamp: float,
                          candle: dict = None, kline: dict = None, init_call: bool = False):
        context = context_management.get_full_context(
            self.trading_mode, matrix_id, cryptocurrency, symbol, time_frame,
            trigger_source, trigger_cache_timestamp, candle, kline, init_call=init_call
        )
        self.last_call_by_timeframe[time_frame] = \
            (matrix_id, cryptocurrency, symbol, time_frame, trigger_source, trigger_cache_timestamp, candle, kline, init_call)
        context.matrix_id = matrix_id
        context.cryptocurrency = cryptocurrency
        context.symbol = symbol
        context.time_frame = time_frame
        initialized = True
        run_data_writer = databases.RunDatabasesProvider.instance().get_run_db(self.exchange_manager.bot_id)
        try:
            await self._pre_script_call(context)
            await self.trading_mode.get_script(live=True)(context)
        except errors.UnreachableExchange:
            raise
        except (commons_errors.MissingDataError, commons_errors.ExecutionAborted) as e:
            self.logger.debug(f"Script execution aborted: {e}")
            initialized = run_data_writer.are_data_initialized
        except Exception as e:
            self.logger.exception(e, True, f"Error when running script: {e}")
        finally:
            if not self.exchange_manager.is_backtesting:
                if context.has_cache(context.symbol, context.time_frame):
                    await context.get_cache().flush()
                for symbol in self.exchange_manager.exchange_config.traded_symbol_pairs:
                    await databases.RunDatabasesProvider.instance().get_symbol_db(
                        self.exchange_manager.bot_id,
                        self.exchange_manager.exchange_name,
                        symbol
                    ).flush()
            run_data_writer.set_initialized_flags(initialized)
            databases.RunDatabasesProvider.instance().get_symbol_db(self.exchange_manager.bot_id,
                                                                  self.exchange_name, symbol)\
                .set_initialized_flags(initialized, (time_frame,))

    async def _pre_script_call(self, context):
        await basic_keywords.set_leverage(context, await basic_keywords.user_select_leverage(context))

    async def post_trigger(self):
        if not self.exchange_manager.is_backtesting:
            # update db after each run only in live mode
            for database in self.all_databases().values():
                if database:
                    try:
                        await database.flush()
                    except Exception as err:
                        self.logger.exception(err, True, f"Error when flushing database: {err}")



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import octobot_commons.errors as errors
import octobot_commons.logging as logging
import octobot_commons.enums as common_enums
import octobot_commons.constants as common_constants
import octobot_commons.configuration as commons_configuration
import octobot_commons.tentacles_management as class_inspector

import octobot_tentacles_manager.api as configurator

import octobot_trading.enums as enums
import octobot_trading.constants as constants
import octobot_trading.modes as modes
import octobot_trading.modes.script_keywords.dsl as dsl


def get_activated_trading_mode(tentacles_setup_config):
    if tentacles_setup_config is not None:
        try:
            trading_modes = [tentacle_class
                             for tentacle_class in configurator.get_activated_tentacles(tentacles_setup_config)
                             if class_inspector.get_deep_class_from_parent_subclasses(tentacle_class,
                                                                                      modes.AbstractTradingMode)]

            if len(trading_modes) > 1:
                raise errors.ConfigTradingError(
                    f"More than one activated trading mode found in your tentacle configuration, "
                    f"please activate only one")

            elif trading_modes:
                trading_mode_class = class_inspector.get_deep_class_from_parent_subclasses(trading_modes[0],
                                                                                           modes.AbstractTradingMode)

                if trading_mode_class is not None:
                    return trading_mode_class
        except ModuleNotFoundError as e:
            logging.get_logger("get_activated_trading_mode").error(f"Error when loading "
                                                                   f"the activated trading mode: {e}")

    raise errors.ConfigTradingError(f"Please ensure your tentacles configuration file is valid and "
                                    f"at least one trading mode is activated")


def should_emit_trading_signals_user_input(trading_mode, inputs: dict):
    trading_mode.UI.user_input(
        common_constants.CONFIG_EMIT_TRADING_SIGNALS, common_enums.UserInputTypes.BOOLEAN, False, inputs,
        title="Emit trading signals on OctoBot cloud for people to follow.",
        order=commons_configuration.UserInput.MAX_ORDER - 2
    )
    trading_mode.UI.user_input(
        common_constants.CONFIG_TRADING_SIGNALS_STRATEGY, common_enums.UserInputTypes.TEXT, trading_mode.get_name(),
        inputs,
        title="Name of the strategy to send signals on.",
        order=commons_configuration.UserInput.MAX_ORDER - 1,
        other_schema_values={"minLength": 0},
        editor_options={
            common_enums.UserInputOtherSchemaValuesTypes.DEPENDENCIES.value: {
                common_constants.CONFIG_EMIT_TRADING_SIGNALS: True
            }
        }
    )


def is_trading_signal_emitter(trading_mode) -> bool:
    """
    :return: True if the mode should be emitting trading signals according to configuration
    """
    try:
        return trading_mode.trading_config[common_constants.CONFIG_EMIT_TRADING_SIGNALS]
    except KeyError:
        return False


def _get_order_amount_title(side):
    return f"Amount per {side} order. {get_order_amount_value_desc()}"


def get_order_amount_value_desc():
    return "To specify the amount per order, " \
        f"use the following syntax: " \
        f"0.1 to trade 0.1 BTC on BTC/USD (amount in base currency); " \
        f"25q to trade 25 USD worth of BTC on BTC/USD (amount in quote currency); " \
        f"2{dsl.QuantityType.PERCENT.value} to trade 2% of the total holdings of the asset; " \
        f"12{dsl.QuantityType.AVAILABLE_PERCENT.value} to trade 12% of the available holdings; " \
        f"5{dsl.QuantityType.CURRENT_SYMBOL_ASSETS_PERCENT.value} to trade 5% of the available " \
           f"holdings associated to the current traded symbol; " \
        f"5{dsl.QuantityType.TRADED_SYMBOLS_ASSETS_PERCENT.value} to trade 5% of the available " \
           f"holdings associated to all configured trading pairs. " \
        f"Leave empty to auto-compute the amount. Checkout the order amounts syntax from trading modes guides " \
        f"for more details."


def user_select_order_amount(trading_mode, inputs: dict, include_buy=True, include_sell=True,
                             buy_dependencies=None, sell_dependencies=None):
    if include_buy:
        trading_mode.UI.user_input(
            constants.CONFIG_BUY_ORDER_AMOUNT, common_enums.UserInputTypes.TEXT, "", inputs,
            title=_get_order_amount_title("buy/entry"),
            other_schema_values={"minLength": 0},
            editor_options={
                common_enums.UserInputOtherSchemaValuesTypes.DEPENDENCIES.value: buy_dependencies
            } if buy_dependencies else None,
        )
    if include_sell:
        trading_mode.UI.user_input(
            constants.CONFIG_SELL_ORDER_AMOUNT, common_enums.UserInputTypes.TEXT, "", inputs,
            title=_get_order_amount_title("sell/exit"),
            other_schema_values={"minLength": 0},
            editor_options={
                common_enums.UserInputOtherSchemaValuesTypes.DEPENDENCIES.value: sell_dependencies
            } if sell_dependencies else None,
        )


def get_user_selected_order_amount(trading_mode, side) -> str:
    try:
        if side is enums.TradeOrderSide.SELL:
            return trading_mode.trading_config[constants.CONFIG_SELL_ORDER_AMOUNT]
        if side is enums.TradeOrderSide.BUY:
            return trading_mode.trading_config[constants.CONFIG_BUY_ORDER_AMOUNT]
    except KeyError:
        return ""
    raise KeyError(f"Unknown side :{side}")



# cython: language_level=3
#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import typing
import octobot_commons.logging as logging
import octobot_commons.constants as commons_constants

import octobot_trading.errors as errors
import octobot_trading.modes as modes
import octobot_trading.exchanges as exchanges
import octobot_trading.constants as constants
import octobot_trading.util as util


class ExchangeBuilder:
    def __init__(self, config, exchange_name):
        self.logger = logging.get_logger(self.__class__.__name__)
        self.config: dict = config
        self.exchange_name: str = exchange_name

        self.exchange_manager: exchanges.ExchangeManager = exchanges.ExchangeManager(self.config, self.exchange_name)

        self._is_using_trading_modes: bool = True
        self._matrix_id: str = None
        self.trading_config_by_trading_mode: dict = None
        self.auto_start_trading_modes: bool = True

    async def build(self):
        """
        Build the exchange
        """
        try:
            await self._build_exchange_manager()
        except Exception as e:
            # stop exchange manager if an exception occurred when building it
            await self.exchange_manager.stop(warning_on_missing_elements=False)
            raise e
        return self.exchange_manager

    def clear(self):
        self.exchange_manager = None

    async def _build_exchange_manager(self):
        trading_mode_class = None

        try:
            if self._is_using_trading_modes:
                trading_mode_class = modes.get_activated_trading_mode(self.exchange_manager.tentacles_setup_config)
                # handle exchange related requirements if the activated trading mode has any
                self._register_trading_modes_requirements(trading_mode_class, self.exchange_manager.tentacles_setup_config)

            self._ensure_exchange_compatibility()
            await self.exchange_manager.initialize()
            # add exchange to be able to use it
            exchanges.Exchanges.instance().add_exchange(self.exchange_manager, self._matrix_id)

            # initialize exchange for trading if not collecting
            if not self.exchange_manager.exchange_only:

                # initialize trader
                if self.exchange_manager.trader is not None:
                    await self._build_trader()

                # create trading modes
                await self._build_trading_modes_if_required(trading_mode_class)

            # add to global exchanges
            self.exchange_manager.update_debug_info()
        except Exception:
            exchanges.Exchanges.instance().del_exchange(
                self.exchange_manager.exchange_name, self.exchange_manager.id, should_warn=False
            )
            raise

    async def _build_trader(self):
        try:
            # check traders activation
            if not util.is_trader_enabled(self.config) and not util.is_trader_simulator_enabled(self.config):
                raise ValueError(f"No trader simulator nor real trader activated on "
                                 f"{self.exchange_manager.exchange_name}")

            await self.exchange_manager.trader.initialize()
        except ValueError as e:
            self.logger.error(f"An error occurred when creating trader : {e}")
            raise e

    def _register_trading_modes_requirements(self, trading_mode_class, tentacles_setup_config):
        self.exchange_manager.is_trading = trading_mode_class.get_is_trading_on_exchange(self.exchange_name,
                                                                                         tentacles_setup_config)
        # take trading modes candles requirements into account if any
        self.config[commons_constants.CONFIG_TENTACLES_REQUIRED_CANDLES_COUNT] = max(
            self.config[commons_constants.CONFIG_TENTACLES_REQUIRED_CANDLES_COUNT],
            modes.get_required_candles_count(trading_mode_class, tentacles_setup_config)
        )

    async def _build_trading_modes_if_required(self, trading_mode_class):
        if self._is_using_trading_modes:
            # self.exchange_manager.trader can be None if neither simulator or real trader has be set
            if self.exchange_manager.is_trading:
                if self.exchange_manager.trader is None:
                    self.logger.warning(f"There wont be any order created on {self.exchange_name}: neither "
                                        f"simulated nor real trader has been activated.")
                else:
                    self.exchange_manager.trading_modes = await self.build_trading_modes(trading_mode_class)
            else:
                self.logger.info(f"{self.exchange_name} exchange is online and won't be trading")

    async def build_trading_modes(self, trading_mode_class):
        try:
            self._ensure_trading_mode_compatibility(trading_mode_class)
            return await modes.create_trading_modes(
                self.config,
                self.exchange_manager,
                trading_mode_class,
                self.exchange_manager.bot_id,
                trading_config_by_trading_mode=self.trading_config_by_trading_mode,
                auto_start=self.auto_start_trading_modes,
            )
        except errors.TradingModeIncompatibility as e:
            raise e
        except Exception as e:
            self.logger.error(f"An error occurred when initializing trading mode : {e}")
            raise e

    def _ensure_exchange_compatibility(self):
        if self.exchange_manager.is_backtesting or self.exchange_manager.exchange_only:
            # allow backtesting and collecting on incompatible exchange types
            return
        # live exchange: ensure the exchange to be created supports the trading type
        supported_exchange_types = exchanges.get_supported_exchange_types(
            self.exchange_manager.exchange_name,
            self.exchange_manager.tentacles_setup_config
        )
        exchange_type = exchanges.get_exchange_type(self.exchange_manager)
        if exchange_type not in supported_exchange_types:
            raise errors.NotSupported(f"{self.exchange_manager.exchange_name} does not support {exchange_type.value}"
                                      f" trading. "
                                      f"Supported exchange types are {[t.value for t in supported_exchange_types]}.")

    def _ensure_trading_mode_compatibility(self, trading_mode_class):
        to_check_exchange_type = exchanges.get_exchange_type(self.exchange_manager)
        if to_check_exchange_type not in trading_mode_class.get_supported_exchange_types():
            self.logger.error(f"{trading_mode_class.get_name()} is not compatible with a {to_check_exchange_type.value} "
                              f"exchange. Activating it might have unexpected effects.")
        # raise errors.TradingModeIncompatibility if we decide we need to prevent the trading mode creation

    """
    Builder methods
    """

    def is_backtesting(self, backtesting_instance):
        self.exchange_manager.is_backtesting = True
        self.exchange_manager.exchange_config.init_backtesting_exchange_config()
        self.exchange_manager.backtesting = backtesting_instance
        return self

    def is_sandboxed(self, sandboxed: bool):
        self.exchange_manager.is_sandboxed = sandboxed
        return self

    def is_simulated(self):
        self.exchange_manager.is_simulated = True
        self.exchange_manager.trader = exchanges.TraderSimulator(self.config, self.exchange_manager)
        return self

    def is_real(self):
        self.exchange_manager.is_simulated = False
        self.exchange_manager.trader = exchanges.Trader(self.config, self.exchange_manager)
        return self

    def is_using_exchange_type(self, exchange_type):
        if exchange_type == commons_constants.CONFIG_EXCHANGE_FUTURE:
            self.is_future(True)
        elif exchange_type == commons_constants.CONFIG_EXCHANGE_MARGIN:
            self.is_margin(True)
        else:
            # Use spot trading as default trading type
            self.is_spot_only(True)
        return self

    def is_margin(self, use_margin=True):
        self.exchange_manager.is_margin = use_margin
        return self

    def is_future(self, use_future=True, future_contract_type=constants.DEFAULT_SYMBOL_CONTRACT_TYPE):
        self.exchange_manager.is_future = use_future
        if self.exchange_manager.is_backtesting:
            self.exchange_manager.exchange_config.backtesting_exchange_config.future_contract_type = \
                future_contract_type
        return self

    def enable_storage(self, enabled):
        self.exchange_manager.enable_storage = enabled
        return self

    def is_spot_only(self, use_spot_only=True):
        self.exchange_manager.is_spot_only = use_spot_only
        return self

    def is_rest_only(self):
        self.exchange_manager.rest_only = True
        return self

    def is_exchange_only(self):
        self.exchange_manager.exchange_only = True
        return self

    def use_cached_markets(self, use_cached_markets: bool):
        self.exchange_manager.use_cached_markets = use_cached_markets
        return self

    def use_market_filter(self, market_filter: typing.Union[None, typing.Callable[[dict], bool]]):
        self.exchange_manager.market_filter = market_filter
        return self

    def is_ignoring_config(self, ignore_config=True):
        self.exchange_manager.ignore_config = ignore_config
        return self

    def is_broker_enabled(self, is_broker_enabled=False):
        self.exchange_manager.is_broker_enabled = is_broker_enabled
        return self

    def is_without_auth(self):
        self.exchange_manager.without_auth = True
        return self

    def is_checking_credentials(self, check_credentials):
        self.exchange_manager.check_credentials = check_credentials
        return self

    def use_tentacles_setup_config(self, tentacles_setup_config):
        self.exchange_manager.tentacles_setup_config = tentacles_setup_config
        return self

    def set_bot_id(self, bot_id):
        self.exchange_manager.bot_id = bot_id
        return self

    def disable_trading_mode(self):
        self._is_using_trading_modes = False
        return self

    def has_matrix(self, matrix_id):
        self._matrix_id = matrix_id
        return self

    def use_trading_config_by_trading_mode(self, trading_config_by_trading_mode):
        self.trading_config_by_trading_mode = trading_config_by_trading_mode
        return self

    def set_auto_start_trading_modes(self, auto_start_trading_modes):
        self.auto_start_trading_modes = auto_start_trading_modes
        return self


def create_exchange_builder_instance(config, exchange_name):
    return ExchangeBuilder(config, exchange_name)



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.

from octobot_trading.api import symbol_data
from octobot_trading.api import trades
from octobot_trading.api import channels
from octobot_trading.api import exchange
from octobot_trading.api import modes
from octobot_trading.api import trader
from octobot_trading.api import portfolio
from octobot_trading.api import profitability
from octobot_trading.api import orders
from octobot_trading.api import contracts
from octobot_trading.api import storage

from octobot_trading.api.symbol_data import (
    get_symbol_data,
    get_symbol_candles_manager,
    get_symbol_historical_candles,
    create_preloaded_candles_manager,
    are_symbol_candles_initialized,
    get_candles_as_list,
    get_candle_as_list,
    has_symbol_klines,
    get_symbol_klines,
    get_symbol_candles_count,
    get_symbol_close_candles,
    get_symbol_open_candles,
    get_symbol_high_candles,
    get_symbol_low_candles,
    get_symbol_volume_candles,
    get_symbol_time_candles,
    create_new_candles_manager,
    force_set_mark_price,
    is_mark_price_initialized,
    get_config_symbols,
)
from octobot_trading.api.trades import (
    get_trade_history,
    get_completed_pnl_history,
    is_executed_trade,
    is_trade_after_or_at,
    get_total_paid_trading_fees,
    get_trade_exchange_name,
    parse_trade_type,
    trade_to_dict,
    get_win_rate,
)
from octobot_trading.api.channels import (
    subscribe_to_ohlcv_channel,
    subscribe_to_trades_channel,
    subscribe_to_order_channel,
)
from octobot_trading.api.exchange import (
    create_exchange_builder,
    get_exchange_configurations_from_exchange_name,
    get_exchange_manager_from_exchange_name_and_id,
    get_ccxt_exchange_available_time_frames,
    get_exchange_available_required_time_frames,
    get_exchange_configuration_from_exchange_id,
    get_exchange_manager_from_exchange_id,
    get_exchange_managers_from_exchange_ids,
    get_trading_exchanges,
    is_exchange_trading,
    get_exchange_manager_id,
    get_exchange_manager_is_sandboxed,
    get_exchange_current_time,
    get_exchange_backtesting_time_window,
    get_exchange_allowed_time_lag,
    get_exchange_id_from_matrix_id,
    get_matrix_id_from_exchange_id,
    get_all_exchange_ids_from_matrix_id,
    get_exchange_configuration_from_exchange,
    get_all_exchange_ids_with_same_matrix_id,
    get_exchange_names,
    get_exchange_ids,
    get_exchange_name,
    get_exchange_type,
    has_only_ohlcv,
    get_is_backtesting,
    get_backtesting_data_files,
    get_backtesting_data_file,
    get_has_websocket,
    supports_websockets,
    is_compatible_account,
    get_new_ccxt_client,
    get_default_exchange_type,
    is_sponsoring,
    is_broker_enabled,
    get_historical_ohlcv,
    get_bot_id,
    get_supported_exchange_types,
    get_trading_pairs,
    get_all_exchange_symbols,
    get_all_exchange_time_frames,
    get_trading_symbols,
    get_trading_timeframes,
    get_watched_timeframes,
    get_relevant_time_frames,
    get_base_currency,
    get_fees,
    get_max_handled_pair_with_time_frame,
    get_currently_handled_pair_with_time_frame,
    get_required_historical_candles_count,
    is_overloaded,
    store_history_in_run_storage,
    get_enabled_exchanges_names,
    get_auto_filled_exchange_names,
    get_exchange_details,
    cancel_ccxt_throttle_task,
    stop_exchange,
)
from octobot_trading.api.modes import (
    get_trading_modes,
    get_trading_mode_symbol,
    get_trading_mode_followed_strategy_signals_identifier,
    get_trading_mode_current_state,
    get_activated_trading_mode,
    create_trading_modes,
    create_trading_mode,
)
from octobot_trading.api.trader import (
    get_trader,
    has_trader,
    is_trader_enabled_in_config_from_exchange_manager,
    is_trader_existing_and_enabled,
    is_trader_enabled,
    is_trader_enabled_in_config,
    is_trader_simulator_enabled_in_config,
    set_trading_enabled,
    is_trader_simulated,
    get_trader_risk,
    set_trader_risk,
    sell_all_everything_for_reference_market,
    sell_currency_for_reference_market,
    get_current_bot_live_id,
)
from octobot_trading.api.portfolio import (
    get_portfolio,
    get_portfolio_historical_values,
    get_portfolio_currency,
    get_origin_portfolio,
    set_simulated_portfolio_initial_config,
    refresh_real_trader_portfolio,
    format_portfolio,
    parse_decimal_portfolio,
    get_draw_down,
    get_coefficient_of_determination,
    get_usd_like_symbol_from_symbols,
    can_convert_symbol_to_usd_like,
    is_usd_like_coin,
    get_global_portfolio_currencies_values,
)
from octobot_trading.api.profitability import (
    get_profitability_stats,
    get_origin_portfolio_value,
    get_current_portfolio_value,
    get_currency_ref_market_value,
    get_current_holdings_values,
    get_current_crypto_currency_value,
    get_reference_market,
    get_initializing_currencies_prices,
)
from octobot_trading.api.orders import (
    get_open_orders,
    get_order_exchange_name,
    order_to_dict,
    parse_order_type,
    parse_order_status,
    get_order_profitability,
    create_order,
    cancel_all_open_orders,
    cancel_all_open_orders_with_currency,
    cancel_order_with_id,
    LOGGER,
)
from octobot_trading.api.positions import (
    get_positions,
    close_position,
)
from octobot_trading.api.contracts import (
    is_inverse_future_contract,
    is_perpetual_future_contract,
    get_pair_contracts,
    is_handled_contract,
)
from octobot_trading.api.storage import (
    clear_trades_storage_history,
    clear_candles_storage_history,
    clear_database_storage_history,
    clear_transactions_storage_history,
    clear_portfolio_storage_history,
    clear_orders_storage_history,
    get_account_type,
    get_account_type_from_run_metadata,
    get_account_type_from_exchange_manager,
)

__all__ = [
    "get_symbol_data",
    "get_symbol_candles_manager",
    "get_symbol_historical_candles",
    "create_preloaded_candles_manager",
    "are_symbol_candles_initialized",
    "get_candles_as_list",
    "get_candle_as_list",
    "has_symbol_klines",
    "get_symbol_klines",
    "get_symbol_candles_count",
    "get_symbol_close_candles",
    "get_symbol_open_candles",
    "get_symbol_high_candles",
    "get_symbol_low_candles",
    "get_symbol_volume_candles",
    "get_symbol_time_candles",
    "create_new_candles_manager",
    "force_set_mark_price",
    "is_mark_price_initialized",
    "get_config_symbols",
    "get_trade_history",
    "get_completed_pnl_history",
    "is_executed_trade",
    "is_trade_after_or_at",
    "get_total_paid_trading_fees",
    "get_trade_exchange_name",
    "parse_trade_type",
    "trade_to_dict",
    "get_win_rate",
    "subscribe_to_ohlcv_channel",
    "subscribe_to_trades_channel",
    "subscribe_to_order_channel",
    "create_exchange_builder",
    "get_exchange_configurations_from_exchange_name",
    "get_exchange_manager_from_exchange_name_and_id",
    "get_ccxt_exchange_available_time_frames",
    "get_exchange_available_required_time_frames",
    "get_exchange_configuration_from_exchange_id",
    "get_exchange_manager_from_exchange_id",
    "get_exchange_managers_from_exchange_ids",
    "get_trading_exchanges",
    "is_exchange_trading",
    "get_exchange_manager_id",
    "get_exchange_manager_is_sandboxed",
    "get_exchange_current_time",
    "get_exchange_backtesting_time_window",
    "get_exchange_allowed_time_lag",
    "get_exchange_id_from_matrix_id",
    "get_matrix_id_from_exchange_id",
    "get_all_exchange_ids_from_matrix_id",
    "get_exchange_configuration_from_exchange",
    "get_all_exchange_ids_with_same_matrix_id",
    "get_exchange_names",
    "get_exchange_ids",
    "get_exchange_name",
    "get_exchange_type",
    "has_only_ohlcv",
    "get_is_backtesting",
    "get_backtesting_data_files",
    "get_backtesting_data_file",
    "get_has_websocket",
    "supports_websockets",
    "is_compatible_account",
    "get_new_ccxt_client",
    "get_default_exchange_type",
    "is_sponsoring",
    "is_broker_enabled",
    "get_historical_ohlcv",
    "get_bot_id",
    "get_supported_exchange_types",
    "get_trading_pairs",
    "get_all_exchange_symbols",
    "get_all_exchange_time_frames",
    "get_trading_symbols",
    "get_trading_timeframes",
    "get_watched_timeframes",
    "get_relevant_time_frames",
    "get_base_currency",
    "get_fees",
    "get_max_handled_pair_with_time_frame",
    "get_currently_handled_pair_with_time_frame",
    "get_required_historical_candles_count",
    "is_overloaded",
    "store_history_in_run_storage",
    "get_enabled_exchanges_names",
    "get_auto_filled_exchange_names",
    "get_exchange_details",
    "cancel_ccxt_throttle_task",
    "stop_exchange",
    "get_trading_modes",
    "get_trading_mode_symbol",
    "get_trading_mode_followed_strategy_signals_identifier",
    "get_trading_mode_current_state",
    "get_activated_trading_mode",
    "create_trading_modes",
    "create_trading_mode",
    "get_trader",
    "has_trader",
    "is_trader_enabled_in_config_from_exchange_manager",
    "is_trader_existing_and_enabled",
    "is_trader_enabled",
    "is_trader_enabled_in_config",
    "is_trader_simulator_enabled_in_config",
    "set_trading_enabled",
    "is_trader_simulated",
    "get_trader_risk",
    "set_trader_risk",
    "sell_all_everything_for_reference_market",
    "sell_currency_for_reference_market",
    "get_current_bot_live_id",
    "get_portfolio",
    "get_portfolio_historical_values",
    "get_portfolio_currency",
    "get_origin_portfolio",
    "set_simulated_portfolio_initial_config",
    "refresh_real_trader_portfolio",
    "get_draw_down",
    "get_coefficient_of_determination",
    "get_usd_like_symbol_from_symbols",
    "can_convert_symbol_to_usd_like",
    "is_usd_like_coin",
    "get_global_portfolio_currencies_values",
    "get_origin_portfolio_value",
    "get_profitability_stats",
    "format_portfolio",
    "parse_decimal_portfolio",
    "get_current_portfolio_value",
    "get_currency_ref_market_value",
    "get_current_holdings_values",
    "get_current_crypto_currency_value",
    "get_reference_market",
    "get_initializing_currencies_prices",
    "get_open_orders",
    "get_order_exchange_name",
    "order_to_dict",
    "parse_order_type",
    "parse_order_status",
    "get_order_profitability",
    "create_order",
    "cancel_all_open_orders",
    "cancel_all_open_orders_with_currency",
    "cancel_order_with_id",
    "get_positions",
    "close_position",
    "is_inverse_future_contract",
    "is_perpetual_future_contract",
    "get_pair_contracts",
    "is_handled_contract",
    "clear_trades_storage_history",
    "clear_candles_storage_history",
    "clear_database_storage_history",
    "get_account_type",
    "get_account_type_from_run_metadata",
    "get_account_type_from_exchange_manager",
    "clear_transactions_storage_history",
    "clear_portfolio_storage_history",
    "clear_orders_storage_history",
]



# pylint: disable=E0611
#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import asyncio
import time
import decimal

import octobot_commons.constants as common_constants
import octobot_commons.enums as common_enums

import octobot_trading.errors as errors
import octobot_trading.constants as constants
import octobot_trading.enums as enums
import octobot_trading.exchange_data.ohlcv.channel.ohlcv as ohlcv_channel
import octobot_trading.exchanges as exchanges


class OHLCVUpdater(ohlcv_channel.OHLCVProducer):
    CHANNEL_NAME = constants.OHLCV_CHANNEL
    OHLCV_LIMIT = 5  # should be < to candle manager's MAX_CANDLES_COUNT
    OHLCV_OLD_LIMIT = constants.DEFAULT_CANDLE_HISTORY_SIZE  # should be <= to candle manager's MAX_CANDLES_COUNT
    OHLCV_ON_ERROR_TIME = 5
    OHLCV_MIN_REFRESH_TIME = 1
    OHLCV_REFRESH_TIME_THRESHOLD = 1.5  # to prevent spamming at candle closing
    OHLCV_MISSING_DATA_REFRESH_RETRY_MAX_DELAY = 30 * common_constants.MINUTE_TO_SECONDS

    OHLCV_INITIALIZATION_TIMEOUT = 60
    OHLCV_INITIALIZATION_RETRY_DELAY = 10

    def __init__(self, channel):
        super().__init__(channel)
        self.tasks = []
        self.is_initialized = False
        self.initialized_candles_by_tf_by_symbol = {}
        self._logged_historical_candles_incompatibility = False

    async def start(self):
        """
        Creates OHLCV refresh tasks
        """
        if self.single_update_task and not self.single_update_task.done():
            await asyncio.wait_for(self.single_update_task, self.OHLCV_INITIALIZATION_TIMEOUT)
        if not self.is_initialized:
            await self._initialize(False)
        if self.channel is not None:
            if self.channel.is_paused:
                await self.pause()
            else:
                self.tasks = [
                    asyncio.create_task(self._candle_update_loop(time_frame, pair))
                    for time_frame in self._get_time_frames()
                    for pair in self._get_traded_pairs()
                    if self._should_maintain_candle(time_frame, pair)
                ]

    def _get_traded_pairs(self):
        return self.channel.exchange_manager.exchange_config.traded_symbol_pairs

    def _get_time_frames(self):
        return self.channel.exchange_manager.exchange_config.available_time_frames

    def _should_maintain_candle(self, time_frame, pair):
        return not (
            exchanges.is_channel_managed_by_websocket(self.channel.exchange_manager, self.CHANNEL_NAME)
            and self.channel.exchange_manager.exchange_web_socket.is_time_frame_supported(time_frame)
        )

    async def fetch_and_push(self):
        return await self._initialize(True)

    async def _initialize(self, push_initialization_candles):
        try:
            initial_candles_data = await asyncio.gather(*[
                self._initialize_candles(time_frame, pair, True)
                for time_frame in self._get_time_frames()
                for pair in self._get_traded_pairs()
            ])
            if push_initialization_candles:
                await self._push_initial_candles(initial_candles_data)
        except Exception as e:
            self.logger.exception(e, True, f"Error while initializing candles: {e}")
        finally:
            self.logger.debug("Candle history initial fetch completed")
            self.is_initialized = True

    def _get_historical_candles_count(self):
        if self.channel.exchange_manager.exchange_config.required_historical_candles_count > 0:
            if self.channel.exchange_manager.exchange_name in constants.FULL_CANDLE_HISTORY_EXCHANGES:
                return self.channel.exchange_manager.exchange_config.required_historical_candles_count
            if not self._logged_historical_candles_incompatibility:
                self.logger.warning(f"Can't initialize the required "
                                    f"{self.channel.exchange_manager.exchange_config.required_historical_candles_count}"
                                    f" historical candles: {self.channel.exchange_manager.exchange_name} is not "
                                    f"supporting large candles history. Using the {self.OHLCV_OLD_LIMIT} "
                                    f"latest candles instead.")
                self._logged_historical_candles_incompatibility = True
        return self.OHLCV_OLD_LIMIT

    async def _get_init_candles(self, time_frame, pair):
        historical_candles_count_limit = self._get_historical_candles_count()
        if historical_candles_count_limit > constants.DEFAULT_CANDLE_HISTORY_SIZE:
            tf_seconds = common_enums.TimeFramesMinutes[time_frame] * common_constants.MINUTE_TO_SECONDS
            end_time = time.time() * common_constants.MSECONDS_TO_SECONDS
            # add 1 to historical_candles_count_limit to fetch the required count (otherwise one is missing)
            start_time = end_time - (historical_candles_count_limit + 1) * tf_seconds * \
                common_constants.MSECONDS_TO_SECONDS
            candles = []
            async for new_candles in exchanges.get_historical_ohlcv(self.channel.exchange_manager, pair,
                                                                    time_frame, start_time, end_time):
                candles += new_candles
            return candles
        candles: list = await self.channel.exchange_manager.exchange \
            .get_symbol_prices(pair, time_frame, limit=self.OHLCV_OLD_LIMIT)
        return candles

    async def _initialize_candles(self, time_frame, pair, should_retry) \
            -> (str, common_enums.TimeFrames, list):
        """
        Manage timeframe OHLCV data refreshing for all pairs
        :return: a tuple with (trading pair, time_frame, fetched candles)
        """
        self._set_initialized(pair, time_frame, False)
        # fetch history
        candles = None
        if self.channel.exchange_manager.exchange_config is None:
            # exchange stopped
            return None
        try:
            candles: list = await self._get_init_candles(time_frame, pair)
        except errors.FailedRequest as e:
            self.logger.warning(str(e))
        if candles and len(candles) > 1:
            self._set_initialized(pair, time_frame, True)
            await self.channel.exchange_manager.get_symbol_data(pair) \
                .handle_candles_update(time_frame, candles[:-1], replace_all=True, partial=False, upsert=False)
            self.logger.debug(f"Candle history loaded for {pair} on {time_frame}")
            self._set_mark_price_from_candle(pair, candles[-1])
            return pair, time_frame, candles
        elif should_retry:
            # When candle history cannot be loaded, retry to load it later
            self.logger.warning(f"Failed to initialize candle history for {pair} on {time_frame}. Retrying in "
                                f"{self.OHLCV_INITIALIZATION_RETRY_DELAY} seconds")
            # retry only once
            await asyncio.sleep(self.OHLCV_INITIALIZATION_RETRY_DELAY)
            return await self._initialize_candles(time_frame, pair, False)
        else:
            self.logger.warning(f"Failed to initialize candle history for {pair} on {time_frame}. Retrying on "
                                f"the next time frame update")
            return None

    def _set_mark_price_from_candle(self, pair, candle):
        # Initialize mark price with last candle close to allow trading low liquidity markets. Those that might
        # take some time to produce a trade and therefore initialize their mark price, which is
        # required to create orders and might block the trading initialization
        price = decimal.Decimal(str(candle[common_enums.PriceIndexes.IND_PRICE_CLOSE.value]))
        self.channel.exchange_manager.get_symbol_data(pair).handle_mark_price_update(
            price,
            enums.MarkPriceSources.TICKER_CLOSE_PRICE.value
        )
        self.channel.exchange_manager.exchange_personal_data.portfolio_manager.portfolio_value_holder.value_converter.\
            update_last_price(pair, price)

    async def _push_initial_candles(self, initial_candles_data):
        self.logger.debug("Pushing completed initialization candles")
        for initial_candles_tuple in initial_candles_data:
            if initial_candles_tuple is not None:
                pair, time_frame, candles = initial_candles_tuple
                await self._push_complete_candles(time_frame, pair, candles)

    async def _push_complete_candles(self, time_frame, pair, candles):
        await self.push(time_frame, pair, candles[:-1], partial=True)  # push only completed candles

    async def _ensure_candles_initialization(self, pair):
        init_coroutines = tuple(
            self._initialize_candles(time_frame, pair, False)
            for time_frame, initialized in self.initialized_candles_by_tf_by_symbol[pair].items()
            if not initialized
        )
        # call gather only if init_coroutines is not empty for optimization purposes
        if init_coroutines:
            await asyncio.gather(*init_coroutines)

    async def _candle_update_loop(self, time_frame, pair):
        self.logger.debug(f"Starting ohlcv updater loop for {pair} on {time_frame}")
        time_frame_seconds: int = common_enums.TimeFramesMinutes[time_frame] * common_constants.MINUTE_TO_SECONDS
        time_frame_sleep: int = time_frame_seconds
        last_candle_timestamp: float = 0
        missing_data_sleep_time = min(int(time_frame_seconds / 6), self.OHLCV_MISSING_DATA_REFRESH_RETRY_MAX_DELAY)

        while not self.should_stop and not self.channel.is_paused:
            try:
                start_update_time = time.time()
                await self._ensure_candles_initialization(pair)
                # skip uninitialized candles
                if self.initialized_candles_by_tf_by_symbol[pair][time_frame]:
                    candles: list = await self.channel.exchange_manager.exchange.get_symbol_prices(
                        pair,
                        time_frame,
                        limit=self.OHLCV_LIMIT)
                    if candles:
                        last_candle: list = candles[-1]
                    else:
                        last_candle: list = []

                    if last_candle and len(candles) > 1:
                        last_candle_timestamp, sleep_time = await self._refresh_current_candle(
                            time_frame, pair, candles, last_candle, last_candle_timestamp, time_frame_sleep
                        )
                        await asyncio.sleep(self._ensure_correct_sleep_time(sleep_time, time_frame_sleep))
                    else:
                        # not enough candles: retry soon
                        self.logger.debug(f"Missing candles in request results for {pair} on {time_frame}, refreshing "
                                          f"in {missing_data_sleep_time} seconds (available candles: {candles}).")
                        await asyncio.sleep(missing_data_sleep_time)
                else:
                    # candles on this time frame have not been initialized: sleep until the next candle update
                    await asyncio.sleep(max(0.0, time_frame_sleep - (time.time() - start_update_time)))
            except errors.FailedRequest as e:
                self.logger.warning(str(e))
                # avoid spamming on disconnected situation
                await asyncio.sleep(constants.DEFAULT_FAILED_REQUEST_RETRY_TIME)
            except errors.NotSupported:
                self.logger.warning(
                    f"{self.channel.exchange_manager.exchange_name} is not supporting updates")
                await self.pause()
            except Exception as e:
                self.logger.exception(e, True, f"Failed to update ohlcv data for {pair} on {time_frame} : {e}")
                await asyncio.sleep(self.OHLCV_ON_ERROR_TIME)

    async def _refresh_current_candle(self, time_frame, pair, candles, last_candle,
                                      last_candle_timestamp, time_frame_sleep):
        current_candle_timestamp: float = last_candle[common_enums.PriceIndexes.IND_PRICE_TIME.value]
        should_sleep_time: float = current_candle_timestamp + time_frame_sleep - time.time()

        # if we're trying to refresh the current candle => useless
        if last_candle_timestamp == current_candle_timestamp:
            if should_sleep_time < 0:
                # up to date candle is not yet available on exchange: retry in a few seconds
                should_sleep_time = self._ensure_correct_sleep_time(
                    self.OHLCV_REFRESH_TIME_THRESHOLD,
                    time_frame_sleep
                )
            else:
                should_sleep_time = self._ensure_correct_sleep_time(
                    should_sleep_time + time_frame_sleep + self.OHLCV_REFRESH_TIME_THRESHOLD,
                    time_frame_sleep
                )
        else:
            # A fresh candle happened
            last_candle_timestamp = current_candle_timestamp
            await self._push_complete_candles(time_frame, pair, candles)
        return last_candle_timestamp, should_sleep_time

    def _ensure_correct_sleep_time(self, sleep_time_candidate, time_frame_sleep):
        if sleep_time_candidate < OHLCVUpdater.OHLCV_MIN_REFRESH_TIME:
            return OHLCVUpdater.OHLCV_MIN_REFRESH_TIME
        elif sleep_time_candidate > time_frame_sleep:
            return time_frame_sleep
        return sleep_time_candidate

    def _set_initialized(self, pair, time_frame, initialized):
        if pair not in self.initialized_candles_by_tf_by_symbol:
            self.initialized_candles_by_tf_by_symbol[pair] = {}
        self.initialized_candles_by_tf_by_symbol[pair][time_frame] = initialized

    async def resume(self) -> None:
        await super().resume()
        if not self.is_running:
            for task in self.tasks:
                task.cancel()
            await self.run()

    # async def config_callback(self, exchange, cryptocurrency, symbols, time_frames):
    #     if symbols:
    #         for pair in symbols:
    #             self.__create_pair_candle_task(pair)
    #             self.logger.info(f"global_data_callback: added {pair}")
    #
    #     if time_frames:
    #         for time_frame in time_frames:
    #             self.__create_time_frame_candle_task(time_frame)
    #             self.logger.info(f"global_data_callback: added {time_frame}")



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.

import octobot_trading.exchanges.connectors.ccxt.ccxt_websocket_connector as ccxt_websocket_connector
import octobot_trading.exchanges.types as exchanges_types
import octobot_tentacles_manager.api as api


class DefaultWebSocketExchange(exchanges_types.WebSocketExchange):
    DEFAULT_CONNECTOR_CLASS = ccxt_websocket_connector.CCXTWebsocketConnector

    @classmethod
    def get_exchange_connector_class(cls, exchange_manager):
        return api.get_class_from_name_with_activated_required_tentacles(
            name=exchange_manager.exchange.get_associated_websocket_exchange_name(),
            tentacles_setup_config=exchange_manager.tentacles_setup_config,
            parent_class=cls.DEFAULT_CONNECTOR_CLASS
        )

    def create_feeds(self):
        try:
            connector = self.websocket_connector(config=self.config, exchange_manager=self.exchange_manager,
                                                 websocket_name=self.websocket_connector.get_name())
            connector.initialize(pairs=self.pairs, time_frames=self.time_frames, channels=self.channels)
            self.websocket_connectors.append(connector)

        except ValueError as e:
            self.logger.exception(e, True, f"Fail to create feed : {e}")



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import contextlib
import typing

import ccxt
import trading_backend

import octobot_commons.logging as logging
import octobot_commons.constants as common_constants
import octobot_commons.enums as common_enums
import octobot_commons.symbols as common_symbols
import octobot_commons.tentacles_management as tentacles_management

import octobot_tentacles_manager.api as api

import octobot_trading.enums as enums
import octobot_trading.constants as constants
import octobot_trading.exchanges.types as exchanges_types
import octobot_trading.exchanges.implementations as exchanges_implementations
import octobot_trading.exchanges.connectors.ccxt.enums as ccxt_enums
import octobot_trading.exchanges.exchange_details as exchange_details
import octobot_trading.exchanges.exchange_builder as exchange_builder


def get_rest_exchange_class(exchange_name, tentacles_setup_config):
    return search_exchange_class_from_exchange_name(exchanges_types.RestExchange, exchange_name, tentacles_setup_config)


def search_exchange_class_from_exchange_name(exchange_class, exchange_name,
                                             tentacles_setup_config, enable_default=False):
    exchange_class = get_exchange_class_from_name(exchange_class, exchange_name,  tentacles_setup_config, enable_default)
    if exchange_class is not None:
        return exchange_class
    if enable_default:
        return None

    logging.get_logger("ExchangeUtil").debug(f"No specific exchange implementation for {exchange_name} found, "
                                             f"using a default one.")
    children_classes = tentacles_management.get_all_classes_from_parent(exchanges_implementations.DefaultRestExchange)
    if children_classes:
        # last one is the most advanced one
        return children_classes[-1]
    # fallback to DefaultRestExchange
    return exchanges_implementations.DefaultRestExchange


def get_exchange_class_from_name(exchange_parent_class, exchange_name, tentacles_setup_config,
                                 enable_default_implementation, strict_name_matching=False):
    for exchange_candidate in tentacles_management.get_all_classes_from_parent(exchange_parent_class):
        try:
            if _is_exchange_candidate_matching(
                exchange_candidate, exchange_name, tentacles_setup_config,
                enable_default_implementation=enable_default_implementation
            ) and (not strict_name_matching or exchange_candidate.get_name() == exchange_name):
                return exchange_candidate
        except NotImplementedError:
            # A subclass of AbstractExchange will raise a NotImplementedError when calling its get_name() method
            # Here we are returning only a subclass that matches the expected name
            # Only Exchange Tentacles are implementing get_name() to specify the related exchange
            # As we are searching for an exchange_type specific subclass
            # We should ignore classes that raises NotImplementedError
            pass
    auto_filled_exchanges = _get_auto_filled_exchanges(tentacles_setup_config)
    if exchange_name in auto_filled_exchanges:
        return auto_filled_exchanges[exchange_name][0]
    return None


def _get_auto_filled_exchanges(tentacles_setup_config):
    auto_filled_exchanges = {}
    for exchange_candidate in _get_auto_filled_exchanges_tentacles():
        if tentacles_setup_config is None:
            # tentacles_setup_config is required for auto-filled exchanges
            continue
        config = api.get_tentacle_config(
            tentacles_setup_config, exchange_candidate
        )
        for exchange_name in exchange_candidate.supported_autofill_exchanges(config):
            auto_filled_exchanges[exchange_name] = (exchange_candidate, config)
    return auto_filled_exchanges


def get_auto_filled_exchange_names(tentacles_setup_config):
    return list(_get_auto_filled_exchanges(tentacles_setup_config))


def _get_auto_filled_exchanges_tentacles():
    return [
        exchange_candidate
        for exchange_candidate in tentacles_management.get_all_classes_from_parent(exchanges_types.RestExchange)
        if exchange_candidate.HAS_FETCHED_DETAILS
    ]


async def get_exchange_details(
    exchange_name, is_autofilled, tentacles_setup_config, aiohttp_session
) -> exchange_details.ExchangeDetails:
    if is_autofilled:
        auto_filled_exchanges = _get_auto_filled_exchanges(tentacles_setup_config)
        if exchange_name in auto_filled_exchanges:
            return await auto_filled_exchanges[exchange_name][0].get_autofilled_exchange_details(
                aiohttp_session, auto_filled_exchanges[exchange_name][1], exchange_name
            )
    try:
        exchange = getattr(ccxt, exchange_name)()
        return exchange_details.ExchangeDetails(
            exchange.id,
            exchange.name,
            exchange.urls[ccxt_enums.ExchangeColumns.WEBSITE.value],
            exchange.urls[ccxt_enums.ExchangeColumns.API.value],
            exchange.urls[ccxt_enums.ExchangeColumns.LOGO_URL.value],
            False,
        )
    except AttributeError as err:
        raise KeyError from err


def _is_exchange_candidate_matching(
    exchange_candidate, exchange_name, tentacles_setup_config, enable_default_implementation=False
):
    return not exchange_candidate.is_simulated_exchange() and \
           (not exchange_candidate.is_default_exchange() or enable_default_implementation) and \
           exchange_candidate.is_supporting_exchange(exchange_name) and \
           (tentacles_setup_config is None or
            api.is_tentacle_activated_in_tentacles_setup_config(tentacles_setup_config,
                                                                exchange_candidate.__name__))


def get_order_side(order_type):
    return enums.TradeOrderSide.BUY.value if order_type in (enums.TraderOrderType.BUY_LIMIT,
                                                            enums.TraderOrderType.BUY_MARKET) \
        else enums.TradeOrderSide.SELL.value


def log_time_sync_error(logger, exchange_name, error, caller_name):
    logger.error(
        f"{_get_time_sync_error_message(exchange_name, caller_name)} Error: {error}")


def _get_docs_url():
    try:
        import octobot.constants
        return octobot.constants.OCTOBOT_DOCS_URL
    except ImportError:
        return "https://www.octobot.cloud/en/guides"


def _get_exchanges_docs_url():
    try:
        import octobot.constants
        return octobot.constants.EXCHANGES_DOCS_URL
    except ImportError:
        return "https://www.octobot.cloud/en/guides/exchanges"


def _get_time_sync_error_message(exchange_name, caller_name):
    return f"Time synchronization error when calling {caller_name} on {exchange_name.capitalize()}. " \
        f"To fix this, please synchronize your computer's clock. See " \
        f"{_get_docs_url()}/octobot-installation/troubleshoot#time-synchronization"


def get_partners_explanation_message():
    return f"More info on partner exchanges on {_get_exchanges_docs_url()}#partner-exchanges---support-octobot"


def _get_minimal_exchange_config(exchange_name, exchange_config):
    return {
        common_constants.CONFIG_EXCHANGES: {
            exchange_name: exchange_config
        },
        common_constants.CONFIG_TRADER: {
            common_constants.CONFIG_ENABLED_OPTION: False
        },
        common_constants.CONFIG_SIMULATOR: {
            common_constants.CONFIG_ENABLED_OPTION: False
        },
        common_constants.CONFIG_TIME_FRAME: [],
        common_constants.CONFIG_CRYPTO_CURRENCIES: []
    }


def get_enabled_exchanges(config):
    return [
        exchange_name
        for exchange_name in config[common_constants.CONFIG_EXCHANGES]
        if config[common_constants.CONFIG_EXCHANGES][exchange_name].get(
                common_constants.CONFIG_ENABLED_OPTION, True
        )
    ]


@contextlib.asynccontextmanager
async def get_local_exchange_manager(
    exchange_name: str, exchange_config: dict, tentacles_setup_config,
    is_sandboxed: bool, ignore_config=False, builder=None, use_cached_markets=True,
    is_broker_enabled: bool = False,
    market_filter: typing.Union[None, typing.Callable[[dict], bool]] = None
):
    exchange_type = exchange_config.get(common_constants.CONFIG_EXCHANGE_TYPE, get_default_exchange_type(exchange_name))
    builder = builder or exchange_builder.ExchangeBuilder(
        _get_minimal_exchange_config(exchange_name, exchange_config),
        exchange_name
    )
    exchange_manager = await builder.use_tentacles_setup_config(tentacles_setup_config) \
        .is_checking_credentials(False) \
        .is_sandboxed(is_sandboxed) \
        .is_using_exchange_type(exchange_type) \
        .is_exchange_only() \
        .is_rest_only() \
        .is_broker_enabled(is_broker_enabled) \
        .use_cached_markets(use_cached_markets) \
        .use_market_filter(market_filter) \
        .is_ignoring_config(ignore_config) \
        .disable_trading_mode() \
        .build()
    try:
        yield exchange_manager
    finally:
        # do not log stopping message
        logger = exchange_manager.exchange.connector.logger
        logger.disable(True)
        builder.clear()
        await exchange_manager.stop(enable_logs=False)
        logger.disable(False)


async def is_compatible_account(exchange_name: str, exchange_config: dict, tentacles_setup_config, is_sandboxed: bool) \
        -> (bool, bool, str):
    """
    Returns details regarding the compatibility of the account given in parameters
    :return: (True if compatible, True if successful login, error explanation if any)
    """
    async with get_local_exchange_manager(
        exchange_name, exchange_config, tentacles_setup_config, is_sandboxed, ignore_config=False
    ) as local_exchange_manager:
        backend = trading_backend.exchange_factory.create_exchange_backend(local_exchange_manager.exchange)
        try:
            is_compatible, error = await backend.is_valid_account(always_check_key_rights=True)
            if not local_exchange_manager.is_spot_only:
                message = f"Future trading on {exchange_name.capitalize()} requires a supporting account. {error}." \
                          f"Please create a new {exchange_name.capitalize()} account to use futures trading. "
                # only ensure compatibility for non spot trading
                return is_compatible, True, message if error else error
            else:
                # auth didn't fail, spot trading is always allowed
                return True, True, None
        except trading_backend.TimeSyncError:
            return False, False, _get_time_sync_error_message(exchange_name, "backend.is_valid_account")
        except trading_backend.ExchangeAuthError:
            message = f"Invalid {exchange_name.capitalize()} authentication details"
            if is_sandboxed:
                message = f"{message}. Warning: exchange sandbox is enabled, " \
                          f"this means that OctoBot is connecting to the testnet/sandbox version of " \
                          f"{exchange_name.capitalize()} to trade and validate your api key. " \
                          f"Disable sandbox in your accounts configuration if this is not intended."
            return False, False, message
        except trading_backend.APIKeyPermissionsError as err:
            return False, False, f"Please update your API Key permissions: {err}"
        except (AttributeError, Exception) as e:
            return True, False, f"Error when loading exchange account: {e}"


async def get_historical_ohlcv(
    local_exchange_manager, symbol, time_frame, start_time, end_time,
    request_retry_timeout=constants.HISTORICAL_CANDLES_FETCH_DEFAULT_TIMEOUT
):
    """
    Async generator, use as follows:
        async for candles in get_historical_ohlcv(exchange_manager, pair, time_frame, start_time, end_time):
            # candles stuff
    WARNING: start_time and end_time are inclusive boundaries and should be milliseconds timestamps
    request_retry_timeout is a timer in seconds to keep retrying to fetch failed candle requests before giving up
    """
    reached_max = False
    time_frame_sec = common_enums.TimeFramesMinutes[time_frame] * common_constants.MINUTE_TO_SECONDS
    exchange_time = local_exchange_manager.exchange.get_exchange_current_time()
    max_theoretical_time = exchange_time - exchange_time % time_frame_sec
    while start_time < end_time and not reached_max:
        candles = await local_exchange_manager.exchange.retry_till_success(
            request_retry_timeout,
            local_exchange_manager.exchange.get_symbol_prices,
            symbol, time_frame, since=int(start_time)
        )
        if candles:
            while candles and candles[-1][common_enums.PriceIndexes.IND_PRICE_TIME.value] * 1000 > end_time:
                candles.pop(-1)
                reached_max = True
            if candles:
                if candles[-1][common_enums.PriceIndexes.IND_PRICE_TIME.value] >= max_theoretical_time:
                    reached_max = True
                yield candles
                start_time = candles[-1][common_enums.PriceIndexes.IND_PRICE_TIME.value] * 1000
                # avoid fetching the last element twice
                start_time += 1
            else:
                reached_max = True
        else:
            reached_max = True


def get_exchange_type(exchange_manager_instance):
    if exchange_manager_instance.is_spot_only:
        return enums.ExchangeTypes.SPOT
    if exchange_manager_instance.is_future:
        return enums.ExchangeTypes.FUTURE
    if exchange_manager_instance.is_margin:
        return enums.ExchangeTypes.MARGIN
    return enums.ExchangeTypes.SPOT


def get_default_exchange_type(exchange_name):
    if exchange_name in constants.DEFAULT_FUTURE_EXCHANGES:
        return common_constants.CONFIG_EXCHANGE_FUTURE
    return common_constants.DEFAULT_EXCHANGE_TYPE


def get_supported_exchange_types(exchange_name, tentacles_setup_config):
    exchange_class = get_exchange_class_from_name(
        exchanges_types.RestExchange, exchange_name, tentacles_setup_config, False, strict_name_matching=True
    )
    if exchange_class is None:
        # default
        return [enums.ExchangeTypes.SPOT]
    return exchange_class.get_supported_exchange_types()


def update_raw_order_from_raw_trade(order_to_update, raw_trade):
    order_to_update[enums.ExchangeConstantsOrderColumns.INFO.value] = raw_trade.get(
        enums.ExchangeConstantsOrderColumns.INFO.value)
    order_to_update[enums.ExchangeConstantsOrderColumns.EXCHANGE_ID.value] = raw_trade.get(
        enums.ExchangeConstantsOrderColumns.ORDER.value)
    order_to_update[enums.ExchangeConstantsOrderColumns.SYMBOL.value] = raw_trade.get(
        enums.ExchangeConstantsOrderColumns.SYMBOL.value)
    order_to_update[enums.ExchangeConstantsOrderColumns.TYPE.value] = raw_trade.get(
        enums.ExchangeConstantsOrderColumns.TYPE.value)
    order_to_update[enums.ExchangeConstantsOrderColumns.AMOUNT.value] = raw_trade.get(
        enums.ExchangeConstantsOrderColumns.AMOUNT.value)
    order_to_update[enums.ExchangeConstantsOrderColumns.DATETIME.value] = raw_trade.get(
        enums.ExchangeConstantsOrderColumns.DATETIME.value)
    order_to_update[enums.ExchangeConstantsOrderColumns.SIDE.value] = raw_trade.get(
        enums.ExchangeConstantsOrderColumns.SIDE.value)
    order_to_update[enums.ExchangeConstantsOrderColumns.TAKER_OR_MAKER.value] = raw_trade.get(
        enums.ExchangeConstantsOrderColumns.TAKER_OR_MAKER.value)
    order_to_update[enums.ExchangeConstantsOrderColumns.PRICE.value] = raw_trade.get(
        enums.ExchangeConstantsOrderColumns.PRICE.value)
    order_to_update[enums.ExchangeConstantsOrderColumns.TIMESTAMP.value] = order_to_update.get(
        enums.ExchangeConstantsOrderColumns.TIMESTAMP.value,
        raw_trade[enums.ExchangeConstantsOrderColumns.TIMESTAMP.value])
    order_to_update[enums.ExchangeConstantsOrderColumns.STATUS.value] = enums.OrderStatus.FILLED.value
    order_to_update[enums.ExchangeConstantsOrderColumns.FILLED.value] = raw_trade.get(
        enums.ExchangeConstantsOrderColumns.AMOUNT.value)
    order_to_update[enums.ExchangeConstantsOrderColumns.COST.value] = raw_trade.get(
        enums.ExchangeConstantsOrderColumns.COST.value)
    order_to_update[enums.ExchangeConstantsOrderColumns.REMAINING.value] = 0
    order_to_update[
        enums.ExchangeConstantsOrderColumns.FEE.value
    ] = raw_trade.get(enums.ExchangeConstantsOrderColumns.FEE.value)
    return order_to_update


def is_missing_trading_fees(raw_order):
    try:
        return (
            raw_order is not None
            and raw_order[enums.ExchangeConstantsOrderColumns.STATUS.value] in (
                    enums.OrderStatus.CLOSED.value,
                    enums.OrderStatus.FILLED.value,
            )
            and (
                raw_order[enums.ExchangeConstantsOrderColumns.FEE.value] is None
                or raw_order[enums.ExchangeConstantsOrderColumns.FEE.value][
                    enums.FeePropertyColumns.EXCHANGE_ORIGINAL_COST.value
                ] is None
            )
        )
    except KeyError:
        return True


def apply_trades_fees(raw_order, raw_trades_by_exchange_order_id):
    exchange_order_id = raw_order[enums.ExchangeConstantsOrderColumns.EXCHANGE_ID.value]
    if exchange_order_id in raw_trades_by_exchange_order_id and raw_trades_by_exchange_order_id[exchange_order_id]:
        order_fee = raw_trades_by_exchange_order_id[exchange_order_id][0][enums.ExchangeConstantsOrderColumns.FEE.value]
        # add each order's trades fee
        for trade in raw_trades_by_exchange_order_id[exchange_order_id][1:]:
            order_fee[enums.FeePropertyColumns.COST.value] += \
                trade[enums.ExchangeConstantsOrderColumns.FEE.value][enums.FeePropertyColumns.COST.value]
            order_fee[enums.FeePropertyColumns.EXCHANGE_ORIGINAL_COST.value] += \
                trade[enums.ExchangeConstantsOrderColumns.FEE.value][
                    enums.FeePropertyColumns.EXCHANGE_ORIGINAL_COST.value]
        raw_order[enums.ExchangeConstantsOrderColumns.FEE.value] = order_fee


def get_common_traded_quote(exchange_manager) -> typing.Union[str, None]:
    quote = None
    for symbol in exchange_manager.exchange_config.traded_symbols:
        if quote is None:
            quote = symbol.quote
        elif quote != symbol.quote:
            return None
    return quote


def get_associated_symbol(exchange_manager, asset: str, target_asset: str) -> (typing.Union[str, None], bool):
    symbol = common_symbols.merge_currencies(asset, target_asset)
    is_reversed_symbol = False
    if symbol not in exchange_manager.client_symbols:
        # try reversed
        reversed_symbol = common_symbols.merge_currencies(target_asset, asset)
        if reversed_symbol not in exchange_manager.client_symbols:
            return None, is_reversed_symbol
        symbol = reversed_symbol
        is_reversed_symbol = True
    return symbol, is_reversed_symbol



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import math
import pytest

from octobot_trading.enums import ExchangeConstantsMarketStatusColumns as Ecmsc
import octobot_trading.personal_data as personal_data
from tests import event_loop

# All test coroutines will be treated as marked.
pytestmark = pytest.mark.asyncio


async def test_adapt_price():
    # will use symbol market
    symbol_market = {Ecmsc.PRECISION.value: {Ecmsc.PRECISION_PRICE.value: 4}}
    assert personal_data.adapt_price(symbol_market, 0.0001) == 0.0001
    assert personal_data.adapt_price(symbol_market, 0.00015) == 0.0001
    assert personal_data.adapt_price(symbol_market, 0.005) == 0.005
    assert personal_data.adapt_price(symbol_market, 1) == 1

    assert personal_data.adapt_price(symbol_market, 56.5128597145) == 56.5128
    assert personal_data.adapt_price(symbol_market, 1251.0000014576121234854513) == 1251.0000

    # will use default (CURRENCY_DEFAULT_MAX_PRICE_DIGITS)
    symbol_market = {Ecmsc.PRECISION.value: {}}
    assert personal_data.adapt_price(symbol_market, 0.0001) == 0.0001
    assert personal_data.adapt_price(symbol_market, 0.00015) == 0.00014999
    assert personal_data.adapt_price(symbol_market, 0.005) == 0.005
    assert personal_data.adapt_price(symbol_market, 1) == 1.0000000000000000000000001
    assert personal_data.adapt_price(symbol_market, 1) == 1

    assert personal_data.adapt_price(symbol_market, 56.5128597145) == 56.51285971
    assert personal_data.adapt_price(symbol_market, 1251.0000014576121234854513) == 1251.00000145


async def test_get_additional_dusts_to_quantity_if_necessary():
    symbol_market = {Ecmsc.LIMITS.value: {
        Ecmsc.LIMITS_AMOUNT.value: {
            Ecmsc.LIMITS_AMOUNT_MIN.value: 0.5
        },
        Ecmsc.LIMITS_COST.value: {
            Ecmsc.LIMITS_COST_MIN.value: 1
        }
    }}

    current_symbol_holding = 5
    quantity = 3
    price = 1
    assert personal_data.add_dusts_to_quantity_if_necessary(quantity,
                                                            price,
                                                            symbol_market,
                                                            current_symbol_holding) == quantity + 0

    current_symbol_holding = 5
    quantity = 4
    price = 1
    assert personal_data.add_dusts_to_quantity_if_necessary(quantity,
                                                            price,
                                                            symbol_market,
                                                            current_symbol_holding) == quantity + 1

    current_symbol_holding = 5
    quantity = 4.5
    price = 1
    assert personal_data.add_dusts_to_quantity_if_necessary(quantity,
                                                            price,
                                                            symbol_market,
                                                            current_symbol_holding) == quantity + 0.5

    symbol_market = {Ecmsc.LIMITS.value: {
        Ecmsc.LIMITS_AMOUNT.value: {
            Ecmsc.LIMITS_AMOUNT_MIN.value: 0.005
        },
        Ecmsc.LIMITS_COST.value: {
            Ecmsc.LIMITS_COST_MIN.value: 0.00005
        }
    }}

    current_symbol_holding = 0.99000000001
    quantity = 0.9
    price = 0.5
    assert personal_data.add_dusts_to_quantity_if_necessary(quantity,
                                                            price,
                                                            symbol_market,
                                                            current_symbol_holding) == quantity + 0

    current_symbol_holding = 0.99000000001
    quantity = 0.0215245845
    price = 0.5
    assert personal_data.add_dusts_to_quantity_if_necessary(quantity,
                                                            price,
                                                            symbol_market,
                                                            current_symbol_holding) == quantity + 0

    current_symbol_holding = 0.99999999
    quantity = 0.99999
    price = 0.5
    assert personal_data.add_dusts_to_quantity_if_necessary(quantity,
                                                            price,
                                                            symbol_market,
                                                            current_symbol_holding) == 0.99999999

    current_symbol_holding = 0.88
    quantity = 0.7055680057024826
    price = 0.0002
    assert personal_data.add_dusts_to_quantity_if_necessary(quantity,
                                                            price,
                                                            symbol_market,
                                                            current_symbol_holding) == 0.88

    # price = 0 => no dust
    assert personal_data.add_dusts_to_quantity_if_necessary(quantity,
                                                            0,
                                                            symbol_market,
                                                            current_symbol_holding) == quantity


async def test_check_and_adapt_order_details_if_necessary():
    symbol_market = {
        Ecmsc.LIMITS.value: {
            Ecmsc.LIMITS_AMOUNT.value: {
                Ecmsc.LIMITS_AMOUNT_MIN.value: 0.5,
                Ecmsc.LIMITS_AMOUNT_MAX.value: 100,
            },
            Ecmsc.LIMITS_COST.value: {
                Ecmsc.LIMITS_COST_MIN.value: 1,
                Ecmsc.LIMITS_COST_MAX.value: 200
            },
            Ecmsc.LIMITS_PRICE.value: {
                Ecmsc.LIMITS_PRICE_MIN.value: 0.5,
                Ecmsc.LIMITS_PRICE_MAX.value: 50
            },
        },
        Ecmsc.PRECISION.value: {
            Ecmsc.PRECISION_PRICE.value: 8,
            Ecmsc.PRECISION_AMOUNT.value: 8
        }
    }

    invalid_cost_symbol_market = {
        Ecmsc.LIMITS.value: {
            Ecmsc.LIMITS_AMOUNT.value: {
                Ecmsc.LIMITS_AMOUNT_MIN.value: 0.5,
                Ecmsc.LIMITS_AMOUNT_MAX.value: 100,
            },
            Ecmsc.LIMITS_COST.value: {
                Ecmsc.LIMITS_COST_MIN.value: None,
                Ecmsc.LIMITS_COST_MAX.value: None
            },
            Ecmsc.LIMITS_PRICE.value: {
                Ecmsc.LIMITS_PRICE_MIN.value: 0.5,
                Ecmsc.LIMITS_PRICE_MAX.value: 50
            },
        },
        Ecmsc.PRECISION.value: {
            Ecmsc.PRECISION_PRICE.value: 8,
            Ecmsc.PRECISION_AMOUNT.value: 8
        }
    }

    # correct min
    quantity = 0.5
    price = 2
    assert personal_data.check_and_adapt_order_details_if_necessary(quantity, price, symbol_market) == [
        (quantity, price)]

    # correct max
    quantity = 100
    price = 2
    assert personal_data.check_and_adapt_order_details_if_necessary(quantity, price, symbol_market) == [
        (quantity, price)]

    # correct
    quantity = 10
    price = 0.6
    assert personal_data.check_and_adapt_order_details_if_necessary(quantity, price, symbol_market) == [
        (quantity, price)]

    # correct
    quantity = 3
    price = 49.9
    assert personal_data.check_and_adapt_order_details_if_necessary(quantity, price, symbol_market) == [
        (quantity, price)]

    # invalid price > but valid cost
    quantity = 1
    price = 100
    assert personal_data.check_and_adapt_order_details_if_necessary(quantity, price, symbol_market) == [
        (quantity, price)]

    # invalid price < but valid cost
    quantity = 10
    price = 0.1
    assert personal_data.check_and_adapt_order_details_if_necessary(quantity, price, symbol_market) == [
        (quantity, price)]

    # invalid price > and invalid cost from exchange => use price => invalid
    quantity = 1
    price = 100
    assert personal_data.check_and_adapt_order_details_if_necessary(quantity, price, invalid_cost_symbol_market) == []

    # invalid price < and invalid cost from exchange => use price => invalid
    quantity = 1
    price = 0.1
    assert personal_data.check_and_adapt_order_details_if_necessary(quantity, price, invalid_cost_symbol_market) == []

    # invalid cost <
    quantity = 0.5
    price = 1
    assert personal_data.check_and_adapt_order_details_if_necessary(quantity, price, symbol_market) == []

    # invalid cost >
    quantity = 10
    price = 49
    assert personal_data.check_and_adapt_order_details_if_necessary(quantity, price, symbol_market) == [
        (1.83673469, 49),
        (4.08163265, 49),
        (4.08163265, 49)]

    # high cost but no max cost => valid
    quantity = 10
    price = 49
    assert personal_data.check_and_adapt_order_details_if_necessary(quantity, price, symbol_market) == [
        (1.83673469, 49),
        (4.08163265, 49),
        (4.08163265, 49)]

    # invalid cost with invalid price >=
    quantity = 10
    price = 50
    assert personal_data.check_and_adapt_order_details_if_necessary(quantity, price, symbol_market) == [(2, 50),
                                                                                                        (4, 50),
                                                                                                        (4, 50)]

    # invalid cost with invalid price > and invalid cost from exchange => use price => invalid
    quantity = 10
    price = 51
    assert personal_data.check_and_adapt_order_details_if_necessary(quantity, price, invalid_cost_symbol_market) == []

    # invalid amount >
    quantity = 200
    price = 5
    assert personal_data.check_and_adapt_order_details_if_necessary(quantity, price, symbol_market) == [(40.0, 5),
                                                                                                        (40.0, 5),
                                                                                                        (40.0, 5),
                                                                                                        (40.0, 5),
                                                                                                        (40.0, 5)]

    # invalid amount <
    quantity = 0.4
    price = 5
    assert personal_data.check_and_adapt_order_details_if_necessary(quantity, price, symbol_market) == []

    symbol_market = {
        Ecmsc.LIMITS.value: {
            Ecmsc.LIMITS_AMOUNT.value: {
                Ecmsc.LIMITS_AMOUNT_MIN.value: 0.0000005,
                Ecmsc.LIMITS_AMOUNT_MAX.value: 100,
            },
            Ecmsc.LIMITS_COST.value: {
                Ecmsc.LIMITS_COST_MIN.value: 0.00000001,
                Ecmsc.LIMITS_COST_MAX.value: 10
            },
            Ecmsc.LIMITS_PRICE.value: {
                Ecmsc.LIMITS_PRICE_MIN.value: 0.000005,
                Ecmsc.LIMITS_PRICE_MAX.value: 50
            },
        },
        Ecmsc.PRECISION.value: {
            Ecmsc.PRECISION_PRICE.value: 8,
            Ecmsc.PRECISION_AMOUNT.value: 8
        }
    }

    # correct quantity
    # to test _adapt_order_quantity_because_quantity
    quantity = 5000
    price = 0.001
    expected = [(100.0, 0.001), (100.0, 0.001), (100.0, 0.001), (100.0, 0.001), (100.0, 0.001), (100.0, 0.001),
                (100.0, 0.001), (100.0, 0.001), (100.0, 0.001), (100.0, 0.001), (100.0, 0.001), (100.0, 0.001),
                (100.0, 0.001), (100.0, 0.001), (100.0, 0.001), (100.0, 0.001), (100.0, 0.001), (100.0, 0.001),
                (100.0, 0.001), (100.0, 0.001), (100.0, 0.001), (100.0, 0.001), (100.0, 0.001), (100.0, 0.001),
                (100.0, 0.001), (100.0, 0.001), (100.0, 0.001), (100.0, 0.001), (100.0, 0.001), (100.0, 0.001),
                (100.0, 0.001), (100.0, 0.001), (100.0, 0.001), (100.0, 0.001), (100.0, 0.001), (100.0, 0.001),
                (100.0, 0.001), (100.0, 0.001), (100.0, 0.001), (100.0, 0.001), (100.0, 0.001), (100.0, 0.001),
                (100.0, 0.001), (100.0, 0.001), (100.0, 0.001), (100.0, 0.001), (100.0, 0.001), (100.0, 0.001),
                (100.0, 0.001), (100.0, 0.001)]
    assert personal_data.check_and_adapt_order_details_if_necessary(quantity, price, symbol_market) == expected

    # price = 0 => no order
    quantity = 10
    price = 0
    assert personal_data.check_and_adapt_order_details_if_necessary(quantity, price, symbol_market) == []

    symbol_market_without_max = {
        Ecmsc.LIMITS.value: {
            Ecmsc.LIMITS_AMOUNT.value: {
                Ecmsc.LIMITS_AMOUNT_MIN.value: 0.0000005,
                Ecmsc.LIMITS_AMOUNT_MAX.value: None,
            },
            Ecmsc.LIMITS_COST.value: {
                Ecmsc.LIMITS_COST_MIN.value: 0.00000001,
                Ecmsc.LIMITS_COST_MAX.value: None
            },
            Ecmsc.LIMITS_PRICE.value: {
                Ecmsc.LIMITS_PRICE_MIN.value: 0.000005,
                Ecmsc.LIMITS_PRICE_MAX.value: None
            },
        },
        Ecmsc.PRECISION.value: {
            Ecmsc.PRECISION_PRICE.value: 8,
            Ecmsc.PRECISION_AMOUNT.value: 8
        }
    }

    # high cost but no max cost => no split
    quantity = 10
    price = 49
    assert personal_data.check_and_adapt_order_details_if_necessary(quantity, price, symbol_market_without_max) == [
        (10, 49)]

    # high quantity but no max quantity => no split
    quantity = 10000000
    price = 49
    assert personal_data.check_and_adapt_order_details_if_necessary(quantity, price, symbol_market_without_max) == \
           [(10000000, 49)]

    # high price but no max price => no split
    quantity = 10
    price = 4900000
    assert personal_data.check_and_adapt_order_details_if_necessary(quantity, price, symbol_market_without_max) == [
        (10, 4900000)]


async def test_split_orders():
    symbol_market = {
        Ecmsc.LIMITS.value: {
            Ecmsc.LIMITS_AMOUNT.value: {
                Ecmsc.LIMITS_AMOUNT_MIN.value: 1,
                Ecmsc.LIMITS_AMOUNT_MAX.value: 100,
            },
            Ecmsc.LIMITS_COST.value: {
                Ecmsc.LIMITS_COST_MIN.value: 1,
                Ecmsc.LIMITS_COST_MAX.value: 30
            },
            Ecmsc.LIMITS_PRICE.value: {
                Ecmsc.LIMITS_PRICE_MIN.value: 0.001,
                Ecmsc.LIMITS_PRICE_MAX.value: 1000
            },
        },
        Ecmsc.PRECISION.value: {
            Ecmsc.PRECISION_PRICE.value: 1,
            Ecmsc.PRECISION_AMOUNT.value: 1
        }
    }

    symbol_market_without_max = {
        Ecmsc.LIMITS.value: {
            Ecmsc.LIMITS_AMOUNT.value: {
                Ecmsc.LIMITS_AMOUNT_MIN.value: 1,
                Ecmsc.LIMITS_AMOUNT_MAX.value: None,
            },
            Ecmsc.LIMITS_COST.value: {
                Ecmsc.LIMITS_COST_MIN.value: 1,
                Ecmsc.LIMITS_COST_MAX.value: None
            },
            Ecmsc.LIMITS_PRICE.value: {
                Ecmsc.LIMITS_PRICE_MIN.value: 1,
                Ecmsc.LIMITS_PRICE_MAX.value: None
            },
        },
        Ecmsc.PRECISION.value: {
            Ecmsc.PRECISION_PRICE.value: 1,
            Ecmsc.PRECISION_AMOUNT.value: 1
        }
    }
    max_cost = symbol_market[Ecmsc.LIMITS.value][Ecmsc.LIMITS_COST.value][Ecmsc.LIMITS_COST_MAX.value]
    max_quantity = symbol_market[Ecmsc.LIMITS.value][Ecmsc.LIMITS_AMOUNT.value][Ecmsc.LIMITS_AMOUNT_MAX.value]

    # normal situation, split because of cost
    total_price = 100
    valid_quantity = 5
    price = 20
    assert personal_data.split_orders(total_price, max_cost, valid_quantity,
                                      max_quantity, price, valid_quantity, symbol_market) \
           == [(0.5, 20), (1.5, 20), (1.5, 20), (1.5, 20)]

    # normal situation, split because of quantity
    total_price = 5.0255
    valid_quantity = 502.55
    price = 0.01
    assert personal_data.split_orders(total_price, max_cost, valid_quantity,
                                      max_quantity, price, valid_quantity, symbol_market) \
           == [(2.5, 0.01), (100, 0.01), (100, 0.01), (100, 0.01), (100, 0.01), (100, 0.01)]

    # missing info situation, split because of cost
    max_quantity = None
    total_price = 100
    valid_quantity = 5
    price = 20
    assert personal_data.split_orders(total_price, max_cost, valid_quantity,
                                      max_quantity, price, valid_quantity, symbol_market_without_max) \
           == [(0.5, 20), (1.5, 20), (1.5, 20), (1.5, 20)]

    # missing info situation, split because of quantity
    max_quantity = symbol_market[Ecmsc.LIMITS.value][Ecmsc.LIMITS_AMOUNT.value][Ecmsc.LIMITS_AMOUNT_MAX.value]
    max_cost = None
    total_price = 5.0255
    valid_quantity = 502.55
    price = 0.01
    assert personal_data.split_orders(total_price, max_cost, valid_quantity,
                                      max_quantity, price, valid_quantity, symbol_market_without_max) \
           == [(2.5, 0.01), (100, 0.01), (100, 0.01), (100, 0.01), (100, 0.01), (100, 0.01)]

    # missing info situation, can't split
    max_quantity = None
    max_cost = None
    total_price = 5.0255
    valid_quantity = 502.55
    price = 0.01
    with pytest.raises(RuntimeError):
        assert personal_data.split_orders(total_price, max_cost, valid_quantity,
                                          max_quantity, price, valid_quantity, symbol_market_without_max)


async def test_adapt_quantity():
    # will use symbol market
    symbol_market = {Ecmsc.PRECISION.value: {Ecmsc.PRECISION_AMOUNT.value: 4}}
    assert personal_data.adapt_quantity(symbol_market, 0.0001) == 0.0001
    assert personal_data.adapt_quantity(symbol_market, 0.00015) == 0.0001
    assert personal_data.adapt_quantity(symbol_market, 0.005) == 0.005
    assert personal_data.adapt_quantity(symbol_market, 1) == 1.0000000000000000000000001
    assert personal_data.adapt_quantity(symbol_market, 1) == 1

    assert personal_data.adapt_quantity(symbol_market, 56.5128597145) == 56.5128
    assert personal_data.adapt_quantity(symbol_market, 1251.0000014576121234854513) == 1251.0000

    # will use default (0)
    symbol_market = {Ecmsc.PRECISION.value: {}}
    assert personal_data.adapt_quantity(symbol_market, 0.0001) == 0
    assert personal_data.adapt_quantity(symbol_market, 0.00015) == 0
    assert personal_data.adapt_quantity(symbol_market, 0.005) == 0
    assert personal_data.adapt_quantity(symbol_market, 1) == 1.0000000000000000000000001
    assert personal_data.adapt_quantity(symbol_market, 1) == 1

    assert personal_data.adapt_quantity(symbol_market, 56.5128597145) == 56
    assert personal_data.adapt_quantity(symbol_market, 1251.0000014576121234854513) == 1251


async def test_trunc_with_n_decimal_digits():
    assert personal_data.trunc_with_n_decimal_digits(1.00000000001, 10) == 1
    assert personal_data.trunc_with_n_decimal_digits(1.00000000001, 11) == 1.00000000001
    assert personal_data.trunc_with_n_decimal_digits(578.000145000156, 3) == 578
    assert personal_data.trunc_with_n_decimal_digits(578.000145000156, 4) == 578.0001
    assert personal_data.trunc_with_n_decimal_digits(578.000145000156, 7) == 578.000145
    assert personal_data.trunc_with_n_decimal_digits(11111111111111578.000145000156, 7) == 11111111111111578.000145
    assert personal_data.trunc_with_n_decimal_digits(578.000145000156, 9) == 578.000145
    assert personal_data.trunc_with_n_decimal_digits(578.000145000156, 10) == 578.0001450001
    assert personal_data.trunc_with_n_decimal_digits(578.000145000156, 12) == 578.000145000156
    assert personal_data.trunc_with_n_decimal_digits(11111111111111578.000145000156,
                                                     12) == 11111111111111578.000145000156
    assert math.isnan(personal_data.trunc_with_n_decimal_digits(math.nan, 12))



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import decimal

import octobot_commons.enums

import octobot_trading.enums
import octobot_trading.exchange_data as exchange_data
import octobot_trading.util as util


def get_symbol_data(exchange_manager, symbol, allow_creation=True) -> exchange_data.ExchangeSymbolData:
    return exchange_manager.exchange_symbols_data.get_exchange_symbol_data(symbol, allow_creation=allow_creation)


def get_symbol_candles_manager(symbol_data, time_frame) -> exchange_data.CandlesManager:
    return symbol_data.symbol_candles[octobot_commons.enums.TimeFrames(time_frame)]


def get_symbol_historical_candles(symbol_data, time_frame, limit=-1) -> object:
    return get_symbol_candles_manager(symbol_data, time_frame).get_symbol_prices(limit)


async def create_preloaded_candles_manager(preloaded_candles):
    candles_manager = exchange_data.PreloadedCandlesManager()
    await candles_manager.initialize()
    candles_manager.replace_all_candles(preloaded_candles)
    return candles_manager


def are_symbol_candles_initialized(exchange_manager, symbol, time_frame) -> bool:
    try:
        return get_symbol_candles_manager(
            get_symbol_data(exchange_manager, symbol, allow_creation=False),
            time_frame
        ).candles_initialized
    except KeyError:
        return False


def get_candles_as_list(candles_arrays) -> list:
    return [
        exchange_data.get_candle_as_list(candles_arrays, index)
        for index in range(len(candles_arrays[0]))
    ]


def get_candle_as_list(candles_arrays, candle_index=0) -> list:
    return exchange_data.get_candle_as_list(candles_arrays, candle_index)


def has_symbol_klines(symbol_data, time_frame) -> bool:
    return octobot_commons.enums.TimeFrames(time_frame) in symbol_data.symbol_klines


def get_symbol_klines(symbol_data, time_frame) -> list:
    return symbol_data.symbol_klines[octobot_commons.enums.TimeFrames(time_frame)].kline


def get_symbol_candles_count(symbol_data, time_frame) -> int:
    return get_symbol_candles_manager(symbol_data, time_frame).get_symbol_candles_count()


def get_symbol_close_candles(symbol_data, time_frame, limit=-1, include_in_construction=False):
    return exchange_data.get_symbol_close_candles(symbol_data, time_frame, limit, include_in_construction)


def get_symbol_open_candles(symbol_data, time_frame, limit=-1, include_in_construction=False):
    return exchange_data.get_symbol_open_candles(symbol_data, time_frame, limit, include_in_construction)


def get_symbol_high_candles(symbol_data, time_frame, limit=-1, include_in_construction=False):
    return exchange_data.get_symbol_high_candles(symbol_data, time_frame, limit, include_in_construction)


def get_symbol_low_candles(symbol_data, time_frame, limit=-1, include_in_construction=False):
    return exchange_data.get_symbol_low_candles(symbol_data, time_frame, limit, include_in_construction)


def get_symbol_volume_candles(symbol_data, time_frame, limit=-1, include_in_construction=False):
    return exchange_data.get_symbol_volume_candles(symbol_data, time_frame, limit, include_in_construction)


def get_symbol_time_candles(symbol_data, time_frame, limit=-1, include_in_construction=False):
    return exchange_data.get_symbol_time_candles(symbol_data, time_frame, limit, include_in_construction)


def create_new_candles_manager(candles=None, max_candles_count=None) -> exchange_data.CandlesManager:
    manager = exchange_data.CandlesManager(max_candles_count=max_candles_count)
    if candles is not None:
        manager.replace_all_candles(candles)
    return manager


def force_set_mark_price(exchange_manager, symbol, price):
    exchange_manager.exchange_symbols_data.get_exchange_symbol_data(symbol).prices_manager.\
        set_mark_price(decimal.Decimal(str(price)), octobot_trading.enums.MarkPriceSources.EXCHANGE_MARK_PRICE.value)


def is_mark_price_initialized(exchange_manager, symbol: str) -> bool:
    return exchange_manager.exchange_symbols_data.get_exchange_symbol_data(symbol).prices_manager.\
        valid_price_received_event.is_set()


def get_config_symbols(config, enabled_only) -> list:
    return util.get_symbols(config, enabled_only)



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import octobot_trading.enums as trading_enums
import octobot_trading.exchanges.types.rest_exchange as rest_exchange


class DefaultRestExchange(rest_exchange.RestExchange):
    @classmethod
    def get_name(cls) -> str:
        return cls.__name__

    @classmethod
    def is_default_exchange(cls) -> bool:
        return True

    @classmethod
    def is_supporting_exchange(cls, exchange_candidate_name) -> bool:
        return isinstance(exchange_candidate_name, str)

    async def switch_to_account(self, account_type: trading_enums.AccountTypes):
        # Currently not supported
        pass

    @classmethod
    def load_user_inputs_from_class(cls, tentacles_setup_config, tentacle_config):
        # default exchange can't be configured
        return {}
